"""Types for configuring components of MADSci"""

from typing import Any, Optional, Union

from madsci.common.types.base_types import BaseModel
from pydantic import Field


class ConfigParameterDefinition(BaseModel, extra="allow"):
    """A configuration parameter definition for a MADSci system component."""

    name: str = Field(
        title="Parameter Name",
        description="The name of the parameter.",
    )
    description: Optional[str] = Field(
        title="Parameter Description",
        description="A description of the parameter.",
        default=None,
    )
    default: Optional[Any] = Field(
        title="Parameter Default",
        description="The default value of the parameter.",
        default=None,
    )
    required: bool = Field(
        title="Parameter Required",
        description="Whether the parameter is required.",
        default=False,
    )


class ConfigParameterWithValue(ConfigParameterDefinition):
    """A configuration parameter definition with value set"""

    value: Optional[Any] = Field(
        title="Parameter Value",
        description="The value of the parameter, if set",
        default=None,
    )


class ConfigNamespaceDefinition(BaseModel, extra="allow"):
    """A namespace for configuration parameters."""

    namespace: str = Field(
        title="Namespace Name",
        description="The name of the namespace.",
    )
    description: Optional[str] = Field(
        title="Namespace Description",
        description="A description of the namespace.",
        default=None,
    )
    parameters: dict[
        str, Union[ConfigParameterDefinition, "ConfigNamespaceDefinition"]
    ] = Field(
        title="Namespace Parameters",
        description="The config parameters in the namespace.",
        default_factory=dict,
    )



"""Types for MADSci Workcell configuration."""

from pathlib import Path
from typing import Annotated, Any, ClassVar, Literal, Optional, Union

from madsci.common.types.base_types import (
    BaseModel,
    Error,
    LoadConfig,
    ModelLink,
    PathLike,
    new_ulid_str,
)
from madsci.common.types.datapoint_types import ObjectStorageDefinition
from madsci.common.types.event_types import EventClientConfig
from madsci.common.types.lab_types import ManagerType
from madsci.common.types.location_types import Location, LocationDefinition
from madsci.common.types.node_types import Node, NodeDefinition
from madsci.common.types.workflow_types import Workflow
from madsci.common.validators import ulid_validator
from pydantic import computed_field
from pydantic.functional_validators import field_validator
from pydantic.networks import AnyUrl
from sqlmodel.main import Field


class WorkcellDefinition(BaseModel, extra="allow"):
    """Configuration for a MADSci Workcell."""

    _definition_file_patterns: ClassVar[list] = [
        "*workcell.yaml",
        "*workcell.yml",
        "*workcell.manager.yml",
        "*workcell.manager.yaml",
    ]
    _definition_cli_flag: ClassVar[list] = [
        "--workcell",
        "--workcell-definition",
        "--definition",
        "--workcell-definition-file",
        "-f",
    ]

    workcell_name: str = Field(
        title="Workcell Name", description="The name of the workcell."
    )
    manager_type: Literal[ManagerType.WORKCELL_MANAGER] = Field(
        title="Manager Type",
        description="The type of manager",
        default=ManagerType.WORKCELL_MANAGER,
    )
    workcell_id: str = Field(
        title="Workcell ID",
        description="The ID of the workcell.",
        default_factory=new_ulid_str,
    )
    description: Optional[str] = Field(
        default=None,
        title="Workcell Description",
        description="A description of the workcell.",
    )
    config: Annotated[
        "WorkcellConfig",
        Field(
            title="Workcell Configuration",
            description="The configuration for the workcell.",
            default_factory=lambda: WorkcellConfig(),
        ),
        LoadConfig(use_fields_as_cli_args=True),
    ]
    nodes: dict[str, Union["NodeDefinition", AnyUrl, PathLike]] = Field(
        default_factory=dict,
        title="Workcell Node URLs",
        description="The URL, path, or definition for each node in the workcell.",
    )
    locations: list[LocationDefinition] = Field(
        default_factory=list,
        title="Workcell Locations",
        description="The Locations used in the workcell.",
    )

    @computed_field
    @property
    def workcell_directory(self) -> Path:
        """The directory for the workcell."""
        return Path(self.config.workcells_directory) / self.workcell_name

    is_ulid = field_validator("workcell_id")(ulid_validator)


class WorkcellLink(ModelLink[WorkcellDefinition]):
    """Link to a MADSci Workcell Definition."""

    definition: Optional[WorkcellDefinition] = Field(
        title="Workcell Definition",
        description="The actual definition of the workcell.",
        default=None,
    )


class WorkcellStatus(BaseModel):
    """Represents the status of a MADSci workcell."""

    paused: bool = Field(
        default=False,
        title="Workcell Paused",
        description="Whether the workcell is paused.",
    )
    errored: bool = Field(
        default=False,
        title="Workcell Errored",
        description="Whether the workcell is in an error state.",
    )
    errors: list[Error] = Field(
        default_factory=list,
        title="Workcell Errors",
        description="A list of errors the workcell has encountered.",
    )
    initializing: bool = Field(
        default=False,
        title="Workcell Initializing",
        description="Whether the workcell is initializing.",
    )
    shutdown: bool = Field(
        default=False,
        title="Workcell Shutdown",
        description="Whether the workcell is shutting down.",
    )
    locked: bool = Field(
        default=False,
        title="Workcell Locked",
        description="Whether the workcell is locked.",
    )

    @computed_field
    @property
    def ok(self) -> bool:
        """Whether the workcell is in a good state."""
        return not any(
            [
                self.paused,
                self.errored,
                self.initializing,
                self.shutdown,
                self.locked,
            ]
        )

    @field_validator("errors", mode="before")
    @classmethod
    def ensure_list_of_errors(cls, v: Any) -> Any:
        """Ensure that errors is a list of MADSci Errors"""
        if isinstance(v, str):
            return [Error(message=v)]
        if isinstance(v, Error):
            return [v]
        if isinstance(v, Exception):
            return [Error.from_exception(v)]
        if isinstance(v, list):
            for i, item in enumerate(v):
                if isinstance(item, str):
                    v[i] = Error(message=item)
                elif isinstance(item, Exception):
                    v[i] = Error.from_exception(item)
        return v


class WorkcellState(BaseModel):
    """Represents the live state of a MADSci workcell."""

    status: WorkcellStatus = Field(
        default_factory=WorkcellStatus,
        title="Workcell Status",
        description="The status of the workcell.",
    )
    workflow_queue: list[Workflow] = Field(
        default_factory=list,
        title="Workflow Queue",
        description="The queue of workflows in non-terminal states.",
    )
    workcell_definition: WorkcellDefinition = Field(
        title="Workcell Definition",
        description="The definition of the workcell.",
    )
    nodes: dict[str, Node] = Field(
        default_factory=dict,
        title="Workcell Nodes",
        description="The nodes in the workcell.",
    )
    locations: dict[str, Location] = Field(
        default_factory=dict,
        title="Workcell Locations",
        description="The locations in the workcell.",
    )


class WorkcellConfig(BaseModel):
    """Configuration for a MADSci Workcell."""

    host: str = Field(
        default="127.0.0.1",
        title="Host",
        description="The host to run the workcell manager on.",
    )
    port: int = Field(
        default=8005,
        title="Port",
        description="The port to run the workcell manager on.",
    )
    workcells_directory: Optional[PathLike] = Field(
        title="Workcells Directory",
        description="Directory used to store workcell-related files in. Defaults to ~/.madsci/workcells. Workcell-related filess will be stored in a sub-folder with the workcell name.",
        default_factory=lambda: Path("~") / ".madsci" / "workcells",
    )
    redis_host: str = Field(
        default="localhost",
        title="Redis Host",
        description="The hostname for the redis server .",
    )
    redis_port: int = Field(
        default=6379,
        title="Redis Port",
        description="The port for the redis server.",
    )
    redis_password: Union[str, None] = Field(
        default=None,
        title="Redis Password",
        description="The password for the redis server.",
    )
    event_client_config: Optional[EventClientConfig] = Field(
        default=None,
        title="Event Client Configuration",
        description="The configuration for a MADSci event client.",
    )
    scheduler_update_interval: float = Field(
        default=2.0,
        title="Scheduler Update Interval",
        description="The interval at which the scheduler runs, in seconds. Must be >= node_update_interval",
    )
    node_update_interval: float = Field(
        default=1.0,
        title="Node Update Interval",
        description="The interval at which the workcell queries its node's states, in seconds.Must be <= scheduler_update_interval",
    )
    auto_start: bool = Field(
        default=True,
        title="Auto Start",
        description="Whether the workcell should automatically create a new Workcell Manager and start it when the lab starts, registering it with the Lab Server.",
    )
    clear_workflows: bool = Field(
        default=False,
        title="Clear Workflows",
        description="Whether the workcell should clear old workflows on restart",
    )
    cold_start_delay: int = Field(
        default=0,
        title="Cold Start Delay",
        description="How long the Workcell engine should sleep on startup",
    )
    scheduler: str = Field(
        default="madsci.workcell_manager.schedulers.default_scheduler",
        title="scheduler",
        description="Scheduler module that contains a Scheduler class that inherits from AbstractScheduler to use",
    )
    data_server_url: Optional[AnyUrl] = Field(
        default=None,
        title="Data Client URL",
        description="The URL for the data client.",
    )
    resource_server_url: Optional[AnyUrl] = Field(
        default=None,
        title="Resource Server URL",
        description="The URL for the resource server.",
    )
    mongo_url: Optional[str] = Field(
        default=None,
        title="MongoDB URL",
        description="The URL for the mongo database.",
    )
    get_action_result_retries: int = Field(
        default=3,
        title="Get Action Result Retries",
        description="Number of times to retry getting an action result",
    )
    minio_client_config: Optional[ObjectStorageDefinition] = Field(
        title="Object Storage Configuration",
        description="Configuration for S3-compatible object storage using MinIO.",
        default=None,
    )



"""Helper functions for manipulating data"""

import copy
import re
from typing import Any


def walk_and_replace(
    input_dict: dict[str, Any], input_parameters: dict[str, Any]
) -> dict[str, Any]:
    """Recursively walk the input dictionary and replace all parameters"""
    updated_dict = copy.deepcopy(input_dict)
    for key, val in input_dict.items():
        if type(val) is str:
            updated_dict[key] = value_substitution(val, input_parameters)
        elif type(input_dict[key]) is dict:
            updated_dict[key] = walk_and_replace(val, input_parameters)
        if type(key) is str:
            new_key = value_substitution(key, input_parameters)
            updated_dict[new_key] = updated_dict[key]
            if key is not new_key:
                updated_dict.pop(key, None)
    return updated_dict


def value_substitution(input_string: str, input_parameters: dict[str, Any]) -> str:
    """Perform $-string and ${}-string substitution on input string, returns string with substituted values"""
    # * Check if the entire string is a simple parameter reference
    if type(input_string) is str and re.match(r"^\$[A-z0-9_\-]*$", input_string):
        if input_string.strip("$") in input_parameters:
            input_string = input_parameters[input_string.strip("$")]
    else:
        # * Replace all parameter references contained in the string
        working_string = input_string
        for match in re.findall(r"((?<!\$)\$(?!\$)[A-z0-9_\-\{]*)(\})", input_string):
            if match[0][1] == "{":
                # * Matches the form ${parameter}
                param_name = match[0].strip("$")
                param_name = param_name.strip("{")
                if param_name in input_parameters:
                    working_string = re.sub(
                        r"((?<!\$)\$(?!\$)[A-z0-9_\-\{]*)(\})",
                        str(input_parameters[param_name]),
                        working_string,
                    )
                    input_string = working_string
            else:
                raise SyntaxError(
                    "Missing opening { in parameter insertion: " + match[0] + "}"
                )
        for match in re.findall(
            r"((?<!\$)\$(?!\$)[A-z0-9_\-]*)(?![A-z0-9_\-])", input_string
        ):
            # * Matches the form $parameter
            param_name = match.strip("$")
            if param_name in input_parameters:
                working_string = re.sub(
                    r"((?<!\$)\$(?!\$)[A-z0-9_\-]*)(?![A-z0-9_\-])",
                    str(input_parameters[param_name]),
                    working_string,
                )
                input_string = working_string
    return input_string



"""Pydantic Models for Resource Definitions, used to define default resources for a node or workcell."""

from typing import Annotated, Any, Literal, Optional, Union

from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.base_types import (
    BaseModel,
    ConfigDict,
    PositiveInt,
    PositiveNumber,
)
from madsci.common.types.event_types import EventClientConfig
from madsci.common.types.lab_types import ManagerDefinition, ManagerType
from madsci.common.types.resource_types.resource_enums import ResourceTypeEnum
from madsci.common.utils import new_name_str
from pydantic import AfterValidator
from pydantic.functional_validators import field_validator
from pydantic.types import Discriminator, Tag
from sqlalchemy.dialects.postgresql import JSON
from sqlmodel import Field


def single_letter_or_digit_validator(value: str) -> str:
    """Validate that the value is a single letter or digit."""
    if not (value.isalpha() and len(value) == 1) or value.isdigit():
        raise ValueError("Value must be a single letter or digit.")
    return value


GridIndex = Union[
    int,
    Annotated[str, AfterValidator(single_letter_or_digit_validator)],
]
GridIndex2D = tuple[GridIndex, GridIndex]
GridIndex3D = tuple[GridIndex, GridIndex, GridIndex]


class ResourceManagerDefinition(ManagerDefinition):
    """Definition for a Resource Manager's Configuration"""

    manager_type: Literal[ManagerType.RESOURCE_MANAGER] = Field(
        title="Manager Type",
        description="The type of the resource manager",
        default=ManagerType.RESOURCE_MANAGER,
    )
    host: str = Field(
        default="127.0.0.1",
        title="Server Host",
        description="The hostname or IP address of the Resource Manager server.",
    )
    port: int = Field(
        default=8003,
        title="Server Port",
        description="The port number of the Resource Manager server.",
    )
    db_url: str = Field(
        default="postgresql://rpl:rpl@localhost:5432/resources",
        title="Database URL",
        description="The URL of the database used by the Resource Manager.",
    )
    event_client_config: Optional[EventClientConfig] = Field(
        default=None,
        title="Event Client Configuration",
        description="Configuration for the event client.",
    )
    custom_types: dict[str, "ResourceDefinitions"] = Field(
        default_factory=dict,
        title="Custom Types",
        description="Custom Types for this resource manager",
    )


class CustomResourceAttributeDefinition(BaseModel, extra="allow"):
    """Definition for a MADSci Custom Resource Attribute."""

    attribute_name: str = Field(
        title="Attribute Name",
        description="The name of the attribute.",
    )
    attribute_description: Optional[str] = Field(
        default=None,
        title="Attribute Description",
        description="A description of the attribute.",
    )
    optional: bool = Field(
        default=False,
        title="Optional",
        description="Whether the attribute is optional.",
    )
    default_value: Any = Field(
        default=None,
        title="Default Value",
        description="The default value of the attribute.",
        sa_type=JSON,
    )


class ResourceDefinition(BaseModel, table=False, extra="allow"):
    """Definition for a MADSci Resource."""

    model_config = ConfigDict(extra="allow")
    resource_name: str = Field(
        title="Resource Name",
        description="The name of the resource.",
        default_factory=new_name_str,
    )

    resource_name_prefix: Optional[str] = Field(
        title="Resource Name Prefix",
        description="A prefix to append the key of the object to for machine instanciated resources",
        default=None,
    )
    resource_class: str = Field(
        title="Resource Class",
        description="The class of the resource. Must match a class defined in the resource manager.",
        default="",
        nullable=False,
    )
    base_type: Literal[ResourceTypeEnum.resource] = Field(
        default=ResourceTypeEnum.resource,
        title="Resource Base Type",
        description="The base type of the resource.",
    )
    resource_description: Optional[str] = Field(
        default=None,
        title="Resource Description",
        description="A description of the resource.",
    )
    owner: OwnershipInfo = Field(
        default_factory=OwnershipInfo,
        title="Ownership Info",
        description="The owner of this resource",
        sa_type=JSON,
    )
    custom_attributes: Optional[list["CustomResourceAttributeDefinition"]] = Field(
        default=None,
        title="Custom Attributes",
        description="Custom attributes used by resources of this type.",
        sa_type=JSON,
    )

    @classmethod
    def discriminate(cls, resource: dict) -> "ResourceDefinition":
        """Discriminate the resource definition based on its base type."""
        from madsci.common.types.resource_types import RESOURCE_TYPE_MAP

        if isinstance(resource, dict):
            resource_type = resource.get("base_type")
        else:
            resource_type = resource.base_type
        return RESOURCE_TYPE_MAP[resource_type]["definition"].model_validate(resource)


class AssetResourceDefinition(ResourceDefinition, table=False):
    """Definition for an asset resource."""

    base_type: Literal[ResourceTypeEnum.asset] = Field(
        default=ResourceTypeEnum.asset,
        title="Resource Base Type",
        description="The base type of the asset.",
    )


class ConsumableResourceDefinition(ResourceDefinition):
    """Definition for a consumable resource."""

    base_type: Literal[ResourceTypeEnum.consumable] = Field(
        default=ResourceTypeEnum.consumable,
        title="Resource Base Type",
        description="The base type of the consumable.",
    )
    unit: Optional[str] = Field(
        default=None,
        title="Resource Unit",
        description="The unit used to measure the quantity of the consumable.",
    )
    quantity: PositiveNumber = Field(
        default=0.0,
        title="Default Resource Quantity",
        description="The initial quantity of the consumable.",
    )
    capacity: Optional[PositiveNumber] = Field(
        default=None,
        title="Resource Capacity",
        description="The initial capacity of the consumable.",
    )


class DiscreteConsumableResourceDefinition(ConsumableResourceDefinition):
    """Definition for a discrete consumable resource."""

    base_type: Literal[ResourceTypeEnum.discrete_consumable] = Field(
        default=ResourceTypeEnum.discrete_consumable,
        title="Resource Base Type",
        description="The base type of the consumable.",
    )
    quantity: PositiveInt = Field(
        default=0,
        title="Default Resource Quantity",
        description="The initial quantity of the consumable.",
    )
    capacity: Optional[PositiveInt] = Field(
        default=None,
        title="Resource Capacity",
        description="The initial capacity of the consumable.",
    )


class ContinuousConsumableResourceDefinition(ConsumableResourceDefinition):
    """Definition for a continuous consumable resource."""

    base_type: Literal[ResourceTypeEnum.continuous_consumable] = Field(
        default=ResourceTypeEnum.continuous_consumable,
        title="Resource Base Type",
        description="The base type of the continuous consumable.",
    )


class ContainerResourceDefinition(ResourceDefinition):
    """Definition for a container resource."""

    base_type: Literal[ResourceTypeEnum.container] = Field(
        default=ResourceTypeEnum.container,
        title="Resource Base Type",
        description="The base type of the container.",
    )
    capacity: Optional[Union[int, float]] = Field(
        default=None,
        title="Container Capacity",
        description="The capacity of the container. If None, uses the type's default_capacity.",
    )
    default_children: Optional[
        Union[list[ResourceDefinition], dict[str, ResourceDefinition]]
    ] = Field(
        default=None,
        title="Default Children",
        description="The default children to create when initializing the container. If None, use the type's default_children.",
    )
    default_child_template: Optional["ResourceDefinitions"] = Field(
        default=None,
        title="Default Child Template",
        description="Template for creating child resources, supporting variable substitution. If None, use the type's default_child_template.",
    )


class CollectionResourceDefinition(ContainerResourceDefinition):
    """Definition for a collection resource. Collections are used for resources that have a number of children, each with a unique key, which can be randomly accessed."""

    base_type: Literal[ResourceTypeEnum.collection] = Field(
        default=ResourceTypeEnum.collection,
        title="Resource Base Type",
        description="The base type of the collection.",
    )
    keys: Optional[Union[int, list[str]]] = Field(
        default=None,
        title="Collection Keys",
        description="The keys for the collection. Can be an integer (converted to 1-based range) or explicit list.",
    )
    default_children: Optional[
        Union[list[ResourceDefinition], dict[str, ResourceDefinition]]
    ] = Field(
        default=None,
        title="Default Children",
        description="The default children to create when initializing the collection. If None, use the type's default_children.",
    )

    @field_validator("keys", mode="before")
    @classmethod
    def validate_keys(cls, v: Union[int, list[str], None]) -> Optional[list[str]]:
        """Convert integer keys to 1-based range if needed."""
        if isinstance(v, int):
            return [str(i) for i in range(1, v + 1)]
        return v


class RowResourceDefinition(ContainerResourceDefinition):
    """Definition for a row resource. Rows are 1D collections of resources. They are treated as single collections (i.e. Collection[Resource])."""

    base_type: Literal[ResourceTypeEnum.row] = Field(
        default=ResourceTypeEnum.row,
        title="Resource Base Type",
        description="The base type of the row.",
    )
    default_children: Optional[dict[str, ResourceDefinition]] = Field(
        default=None,
        title="Default Children",
        description="The default children to create when initializing the collection. If None, use the type's default_children.",
    )
    fill: bool = Field(
        default=False,
        title="Fill",
        description="Whether to populate every empty key with a default child",
    )
    columns: int = Field(
        title="Number of Columns",
        description="The number of columns in the row.",
        ge=0,
    )
    is_one_indexed: bool = Field(
        title="One Indexed",
        description="Whether the numeric index of the object start at 0 or 1",
        default=True,
    )


class GridResourceDefinition(RowResourceDefinition):
    """Definition for a grid resource. Grids are 2D grids of resources. They are treated as nested collections (i.e. Collection[Collection[Resource]])."""

    base_type: Literal[ResourceTypeEnum.grid] = Field(
        default=ResourceTypeEnum.grid,
        title="Resource Base Type",
        description="The base type of the grid.",
    )
    default_children: Optional[dict[str, dict[str, ResourceDefinition]]] = Field(
        default=None,
        title="Default Children",
        description="The default children to create when initializing the collection. If None, use the type's default_children.",
    )
    rows: int = Field(
        default=None,
        title="Number of Rows",
        description="The number of rows in the grid. If None, use the type's rows.",
    )


class VoxelGridResourceDefinition(GridResourceDefinition):
    """Definition for a voxel grid resource. Voxel grids are 3D grids of resources. They are treated as nested collections (i.e. Collection[Collection[Collection[Resource]]])."""

    base_type: Literal[ResourceTypeEnum.voxel_grid] = Field(
        default=ResourceTypeEnum.voxel_grid,
        title="Resource Base Type",
        description="The base type of the voxel grid.",
    )
    default_children: Optional[dict[str, dict[str, dict[str, ResourceDefinition]]]] = (
        Field(
            default=None,
            title="Default Children",
            description="The default children to create when initializing the collection. If None, use the type's default_children.",
        )
    )
    layers: int = Field(
        title="Number of Layers",
        description="The number of layers in the voxel grid. If None, use the type's layers.",
    )

    def get_all_keys(self) -> list:
        """get all keys of this object"""
        return [
            GridIndex3D((i, j, k))
            for i in range(self.columns)
            for j in range(self.rows)
            for k in range(self.layers)
        ]


class SlotResourceDefinition(ContainerResourceDefinition):
    """Definition for a slot resource."""

    base_type: Literal[ResourceTypeEnum.slot] = Field(
        default=ResourceTypeEnum.slot,
        title="Resource Base Type",
        description="The base type of the slot.",
    )

    default_child_quantity: Optional[int] = Field(
        default=None,
        title="Default Child Quantity",
        description="The number of children to create by default. If None, use the type's default_child_quantity.",
        ge=0,
        le=1,
    )
    capacity: Literal[1] = Field(
        title="Capacity",
        description="The capacity of the slot.",
        default=1,
        const=1,
    )


class StackResourceDefinition(ContainerResourceDefinition):
    """Definition for a stack resource."""

    base_type: Literal[ResourceTypeEnum.stack] = Field(
        default=ResourceTypeEnum.stack,
        title="Resource Base Type",
        description="The base type of the stack.",
    )
    default_child_quantity: Optional[int] = Field(
        default=None,
        title="Default Child Quantity",
        description="The number of children to create by default. If None, use the type's default_child_quantity.",
    )


class QueueResourceDefinition(ContainerResourceDefinition):
    """Definition for a queue resource."""

    base_type: Literal[ResourceTypeEnum.queue] = Field(
        default=ResourceTypeEnum.queue,
        title="Resource Base Type",
        description="The base type of the queue.",
    )
    default_child_quantity: Optional[int] = Field(
        default=None,
        title="Default Child Quantity",
        description="The number of children to create by default. If None, use the type's default_child_quantity.",
    )


class PoolResourceDefinition(ContainerResourceDefinition):
    """Definition for a pool resource. Pool resources are collections of consumables with no structure (used for wells, reservoirs, etc.)."""

    base_type: Literal[ResourceTypeEnum.pool] = Field(
        default=ResourceTypeEnum.pool,
        title="Resource Base Type",
        description="The base type of the pool.",
    )
    capacity: Optional[PositiveNumber] = Field(
        title="Capacity",
        description="The default capacity of the pool as a whole.",
        default=None,
    )
    unit: Optional[str] = Field(
        default=None,
        title="Resource Unit",
        description="The unit used to measure the quantity of the pool.",
    )


ResourceDefinitions = Annotated[
    Union[
        Annotated[ResourceDefinition, Tag("resource")],
        Annotated[AssetResourceDefinition, Tag("asset")],
        Annotated[ContainerResourceDefinition, Tag("container")],
        Annotated[CollectionResourceDefinition, Tag("collection")],
        Annotated[RowResourceDefinition, Tag("row")],
        Annotated[GridResourceDefinition, Tag("grid")],
        Annotated[VoxelGridResourceDefinition, Tag("voxel_grid")],
        Annotated[StackResourceDefinition, Tag("stack")],
        Annotated[QueueResourceDefinition, Tag("queue")],
        Annotated[PoolResourceDefinition, Tag("pool")],
        Annotated[SlotResourceDefinition, Tag("slot")],
        Annotated[ConsumableResourceDefinition, Tag("consumable")],
        Annotated[DiscreteConsumableResourceDefinition, Tag("discrete_consumable")],
        Annotated[ContinuousConsumableResourceDefinition, Tag("continuous_consumable")],
    ],
    Discriminator("base_type"),
]



"""Types related to MADSci Resources."""

import re
import string
from typing import Annotated, Any, Literal, Optional, Union

from madsci.common.types.base_types import PositiveInt, PositiveNumber, new_ulid_str
from madsci.common.types.resource_types.definitions import (
    AssetResourceDefinition,
    CollectionResourceDefinition,
    ConsumableResourceDefinition,
    ContainerResourceDefinition,
    ContinuousConsumableResourceDefinition,
    CustomResourceAttributeDefinition,  # noqa: F401
    DiscreteConsumableResourceDefinition,
    GridIndex,
    GridIndex2D,
    GridIndex3D,
    GridResourceDefinition,
    PoolResourceDefinition,
    QueueResourceDefinition,
    ResourceDefinition,
    RowResourceDefinition,
    SlotResourceDefinition,
    StackResourceDefinition,
    VoxelGridResourceDefinition,
)
from madsci.common.types.resource_types.resource_enums import (
    AssetTypeEnum,
    ConsumableTypeEnum,
    ContainerTypeEnum,
    ResourceTypeEnum,
)
from madsci.common.validators import ulid_validator
from pydantic import (
    AnyUrl,
    computed_field,
    model_validator,
)
from pydantic.functional_validators import field_validator
from pydantic.types import Discriminator, Tag, datetime
from sqlalchemy.dialects.postgresql import JSON
from sqlalchemy.sql.sqltypes import String
from sqlmodel import Field
from typing_extensions import Self  # type: ignore


class Resource(ResourceDefinition, extra="allow", table=False):
    """Base class for all MADSci Resources. Used to track any resource that isn't well-modeled by a more specific type."""

    resource_id: str = Field(
        title="Resource ID",
        description="The ID of the resource.",
        nullable=False,
        default_factory=new_ulid_str,
        primary_key=True,
        sa_type=String,
    )
    resource_url: Optional[AnyUrl] = Field(
        title="Resource URL",
        description="The URL of the resource.",
        nullable=True,
        default=None,
        sa_type=String,
    )
    base_type: Literal[ResourceTypeEnum.resource] = Field(
        title="Resource Base Type",
        description="The base type of the resource.",
        nullable=False,
        default=ResourceTypeEnum.resource,
        sa_type=String,
    )
    parent_id: Optional[str] = Field(
        default=None,
        title="Parent Resource",
        description="The parent resource ID, if any.",
    )
    key: Optional[str] = Field(
        default=None,
        title="Key",
        description="The key of the resource in the parent container, if any.",
    )
    attributes: dict = Field(
        default_factory=dict,
        sa_type=JSON,
        title="Attributes",
        description="Custom attributes for the asset.",
    )
    created_at: Optional[datetime] = Field(
        title="Created Datetime",
        description="The timestamp of when the resource was created.",
        default=None,
    )
    updated_at: Optional[datetime] = Field(
        title="Updated Datetime",
        description="The timestamp of when the resource was last updated.",
        default=None,
    )
    removed: bool = Field(
        title="Removed",
        description="Whether the resource has been removed from the lab.",
        nullable=False,
        default=False,
    )
    is_ulid = field_validator("resource_id")(ulid_validator)

    @classmethod
    def discriminate(cls, resource: dict) -> "Resource":
        """Discriminate the resource based on its base type."""
        if isinstance(resource, dict):
            resource_type = resource.get("base_type")
        else:
            resource_type = resource.base_type
        return RESOURCE_TYPE_MAP[resource_type]["model"].model_validate(resource)


class Asset(Resource):
    """Base class for all MADSci Assets. These are tracked resources that aren't consumed (things like samples, labware, etc.)."""

    base_type: Literal[AssetTypeEnum.asset] = Field(
        title="Asset Base Type",
        description="The base type of the asset.",
        nullable=False,
        default=AssetTypeEnum.asset,
    )


class Consumable(Resource):
    """Base class for all MADSci Consumables. These are resources that are consumed (things like reagents, pipette tips, etc.)."""

    base_type: Literal[ConsumableTypeEnum.consumable] = Field(
        title="Consumable Base Type",
        description="The base type of the consumable.",
        nullable=False,
        default=ConsumableTypeEnum.consumable,
    )
    quantity: PositiveNumber = Field(
        title="Quantity",
        description="The quantity of the consumable.",
        default=0,
    )
    capacity: Optional[PositiveNumber] = Field(
        title="Capacity",
        description="The maximum capacity of the consumable.",
        default=None,
    )
    unit: Optional[str] = Field(
        default=None,
        title="Resource Unit",
        description="The unit used to measure the quantity of the consumable.",
    )

    @model_validator(mode="after")
    def validate_consumable_quantity(self) -> Self:
        """Validate that the quantity is less than or equal to the capacity."""
        if self.capacity is not None and self.quantity > self.capacity:
            raise ValueError("Quantity cannot be greater than capacity.")
        return self


class DiscreteConsumable(Consumable):
    """Base class for all MADSci Discrete Consumables. These are consumables that are counted in whole numbers (things like pipette tips, tubes, etc.)."""

    base_type: Literal[ConsumableTypeEnum.discrete_consumable] = Field(
        title="Consumable Base Type",
        description="The base type of the discrete consumable.",
        default=ConsumableTypeEnum.discrete_consumable,
        const=True,
    )
    quantity: PositiveInt = Field(
        title="Quantity",
        description="The quantity of the discrete consumable.",
        default=0,
    )
    capacity: Optional[PositiveInt] = Field(
        title="Capacity",
        description="The maximum capacity of the discrete consumable.",
        default=None,
    )


class ContinuousConsumable(Consumable):
    """Base class for all MADSci Continuous Consumables. These are consumables that are measured in continuous quantities (things like liquids, powders, etc.)."""

    base_type: Literal[ConsumableTypeEnum.continuous_consumable] = Field(
        title="Consumable Base Type",
        description="The base type of the continuous consumable.",
        default=ConsumableTypeEnum.continuous_consumable,
        const=True,
    )
    quantity: PositiveNumber = Field(
        title="Quantity",
        description="The quantity of the continuous consumable.",
        default=0.0,
    )
    capacity: Optional[PositiveNumber] = Field(
        title="Capacity",
        description="The maximum capacity of the continuous consumable.",
        default=None,
    )


class Container(Asset):
    """Data Model for a Container. A container is a resource that can hold other resources."""

    base_type: Literal[ContainerTypeEnum.container] = Field(
        title="Container Base Type",
        description="The base type of the container.",
        nullable=False,
        default=ContainerTypeEnum.container,
    )
    capacity: Optional[PositiveInt] = Field(
        title="Capacity",
        description="The capacity of the container.",
        default=None,
    )
    children: dict[str, "ResourceDataModels"] = Field(
        title="Children",
        description="The children of the container.",
        default_factory=dict,
    )

    def get_child(self, key: str) -> Optional["ResourceDataModels"]:
        """Get a child from the container."""
        return self.children.get(key, None)

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        return len(self.children)

    def extract_children(self) -> dict[str, "ResourceDataModels"]:
        """Extract the children from the container as a flat dictionary."""
        return self.children

    def populate_children(self, children: dict[str, "ResourceDataModels"]) -> None:
        """Populate the children of the container."""
        self.children = children

    @model_validator(mode="after")
    def validate_container_quantity(self) -> Self:
        """Validate that the quantity is less than or equal to the capacity."""
        if self.capacity is not None and self.quantity > self.capacity:
            raise ValueError("Quantity cannot be greater than capacity.")
        return self


class Collection(Container):
    """Data Model for a Collection. A collection is a container that can hold other resources, and which supports random access."""

    base_type: Literal[ContainerTypeEnum.collection] = Field(
        title="Container Base Type",
        description="The base type of the collection.",
        default=ContainerTypeEnum.collection,
        const=True,
    )
    children: dict[str, "ResourceDataModels"] = Field(
        title="Children",
        description="The children of the collection.",
        default_factory=dict,
    )

    def get_child(self, key: str) -> Optional["ResourceDataModels"]:
        """Get a child from the container."""
        return self.children.get(key, None)

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        return len(self.children)

    def extract_children(self) -> dict[str, "ResourceDataModels"]:
        """Extract the children from the collection as a flat dictionary."""
        return self.children

    def populate_children(self, children: dict[str, "ResourceDataModels"]) -> None:
        """Populate the children of the collection."""
        self.children = children


class Row(Container):
    """Data Model for a Row. A row is a container that can hold other resources in a single dimension and supports random access. For example, a row of tubes in a rack or a single-row microplate. Rows are indexed by integers or letters."""

    base_type: Literal[ContainerTypeEnum.row] = Field(
        title="Container Base Type",
        description="The base type of the row.",
        default=ContainerTypeEnum.row,
        const=True,
    )
    columns: int = Field(
        title="Number of Columns",
        description="The number of columns in the row.",
        ge=0,
    )

    children: list[Union["ResourceDataModels", None]] = Field(
        title="Children", description="The children of the row container.", default=[]
    )

    is_one_indexed: bool = Field(
        title="One Indexed",
        description="Whether the numeric index of the object start at 0 or 1, only used with string keys",
        default=True,
    )

    @model_validator(mode="after")
    def set_list(self) -> "Row":
        """populates the children list with none values"""
        if self.children == []:
            self.children = [None] * self.columns
        return self

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        quantity = 0
        for child in self.children:
            if getattr(child, "quantity", None) is not None:
                quantity += child.quantity
        return quantity

    def extract_children(self) -> list["ResourceDataModels"]:
        """return all children"""
        return_dict = {}
        for index, item in enumerate(self.children):
            return_dict[str(index)] = item
        return return_dict

    def get_child(self, key: GridIndex) -> Optional["ResourceDataModels"]:
        """Get a child from the Row."""
        if isinstance(key, str):
            return self.children[int(key) - self.is_one_indexed]
        return self.children[key]

    def numericize_index(self, key: Union[str, GridIndex]) -> GridIndex:
        """Convert a key to a numeric value."""
        if isinstance(key, int):
            return int(key)
        if key.isdigit():
            return int(key) - self.is_one_indexed
        return string.ascii_lowercase.index(key.lower())

    def set_child(self, key: GridIndex, value: "ResourceDataModels") -> None:
        """set a child using a string or int"""
        if isinstance(key, str):
            key = self.numericize_index(key)
        self.children[key] = value
        value.key = str(key)

    def __getitem__(self, index: GridIndex) -> Resource:
        """retrieve a child using an alphanumeric key, if index is a string, uses self.is_one_indexed, otherwise uses normal array rules"""
        return self.get_child(index)

    def __setitem__(self, key: GridIndex, value: Any) -> None:
        """set an item using an alphanumeric key"""
        self.set_child(key, value)

    def check_key_bounds(self, key: Union[str, GridIndex]) -> bool:
        """Check if the key is within the bounds of the grid."""
        return not (int(key) < 0 or int(key) >= self.columns)

    def get_all_keys(self) -> list:
        """get all keys of this object"""
        return list(range(self.columns))

    def populate_children(
        self, children: dict[GridIndex, "ResourceDataModels"]
    ) -> None:
        """Populate the children of the grid."""
        for key, value in children.items():
            self[int(key)] = value


class Grid(Row):
    """Data Model for a Grid. A grid is a container that can hold other resources in two dimensions and supports random access. For example, a 96-well microplate. Grids are indexed by integers or letters."""

    base_type: Literal[ContainerTypeEnum.grid] = Field(
        title="Container Base Type",
        description="The base type of the grid.",
        default=ContainerTypeEnum.grid,
        const=True,
    )
    rows: int = Field(
        title="Number of Rows",
        description="The number of rows in the grid.",
        ge=0,
    )

    children: list[Optional[Row]] = Field(
        title="Children", description="The children of the grid container.", default=[]
    )

    @model_validator(mode="after")
    def initialize_grid(self) -> None:
        """Creates a grid of the correct dimensions"""
        for i in range(self.rows):
            if i >= len(self.children):
                self.children.append(
                    Row(
                        columns=self.columns,
                        resource_name=self.resource_name + "_row_" + str(i),
                    )
                )
            if self.children[i] is None:
                self.children[i] = Row(
                    columns=self.columns,
                    resource_name=self.resource_name + "_row_" + str(i),
                )
        return self

    def split_index(self, key: str) -> GridIndex2D:
        """split an alphanumeric index string into a grid index tuple, uses is_one_indexed for the numerical index"""
        match = re.search(r"(\d)", key)
        if match:
            index = match.start()
            return (
                self.numericize_index(key[:index]),
                self.numericize_index(key[index:]),
            )
        raise (ValueError("Key is in the wrong format!"))

    def get_all_keys(self) -> list:
        """get all keys of this object"""
        return [
            GridIndex2D(i, j) for i in range(self.columns) for j in range(self.rows)
        ]

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        quantity = 0
        for row in self.children:
            if getattr(row, "quantity", None) is not None:
                quantity += row.quantity
        return quantity

    def get_child(
        self, key: Union[str, GridIndex2D, int]
    ) -> Optional["ResourceDataModels"]:
        """Get a child from the Grid."""
        if isinstance(key, str):
            key = self.split_index(key)
        if isinstance(key, int):
            return self.children[key]
        row = self.children[key[0]]
        return row[key[1]]

    def set_child(
        self, key: Union[str, GridIndex2D, int], child: "ResourceDataModels"
    ) -> None:
        """Get a child from the Grid."""
        if isinstance(key, int):
            self.children[key] = child
        else:
            if isinstance(key, str):
                key = self.split_index(key)
            row = self.children[key[0]]
            row[key[1]] = child
            child.key = str(key[1])

    def __getitem__(self, key: Union[str, GridIndex2D, int]) -> Resource:
        """get an item using alphanumeric keys"""

        return self.get_child(key)

    def __setitem__(self, key: str, value: Any) -> None:
        """set an item using alphanumeric keys"""
        self.set_child(key, value)

    def check_key_bounds(self, key: Union[str, GridIndex2D]) -> bool:
        """Check if the key is within the bounds of the grid."""
        if isinstance(key, str):
            key = self.split_index(key)
        elif not isinstance(key, tuple) or len(key) != 2:
            raise ValueError("Key must be a string or a 2-tuple.")
        return not (key[0] < 0 or key[0] >= self.rows) and not (
            key[1] < 0 or key[1] >= self.columns
        )


class VoxelGrid(Grid):
    """Data Model for a Voxel Grid. A voxel grid is a container that can hold other resources in three dimensions and supports random access. Voxel grids are indexed by integers or letters."""

    base_type: Literal[ContainerTypeEnum.voxel_grid] = Field(
        title="Container Base Type",
        description="The base type of the voxel grid.",
        default=ContainerTypeEnum.voxel_grid,
        const=True,
    )
    layers: int = Field(
        title="Number of Layers",
        description="The number of layers in the voxel grid.",
        ge=0,
    )
    children: list[Optional[Grid]] = Field(
        title="Children",
        description="The children of the voxel grid container.",
        default=[],
    )

    @model_validator(mode="after")
    def initialize_grid(self) -> None:
        """Creates a voxel grid of the correct dimension"""
        for i in range(self.layers):
            if i >= len(self.children):
                self.children.append(
                    Grid(
                        resource_name=self.resource_name + "_layer_" + str(i),
                        rows=self.rows,
                        columns=self.columns,
                        children=[
                            Row(
                                columns=self.columns,
                                resource_name=self.resource_name
                                + "_layer_"
                                + str(i)
                                + "_row_"
                                + str(j),
                            )
                            for j in range(self.rows)
                        ],
                    )
                )
            if self.children[i] is None:
                self.children[i] = Grid(
                    resource_name=self.resource_name + "_layer_" + str(i),
                    rows=self.rows,
                    columns=self.columns,
                    children=[
                        Row(
                            columns=self.columns,
                            resource_name=self.resource_name
                            + "_layer_"
                            + str(i)
                            + "_row_"
                            + str(j),
                        )
                        for j in range(self.rows)
                    ],
                )
        return self

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        quantity = 0
        for grid_value in self:
            if getattr(grid_value, "quantity", None) is not None:
                quantity += grid_value.quantity
        return quantity

    def get_child(self, key: GridIndex3D) -> Optional["ResourceDataModels"]:
        """Get a child from the Voxel Grid."""
        grid = self.children[key[2]]
        return grid[(key[0], key[1])]

    def set_child(
        self, key: Union[GridIndex3D, int], child: "ResourceDataModels"
    ) -> None:
        """Get a child from the Grid."""
        if isinstance(key, int):
            self.children[key] = child
        else:
            grid = self.children[key[2]]
            if not isinstance(grid, Grid):
                grid = Grid(
                    resource_name=self.resource_name + "_layer_" + str(key[2]),
                    rows=self.rows,
                    columns=self.columns,
                    children=[
                        Row(
                            columns=self.columns,
                            resource_name=self.resource_name
                            + "_layer_"
                            + str(key[2])
                            + "_row_"
                            + str(j),
                        )
                        for j in range(self.rows)
                    ],
                )
                self.children[key[2]] = grid
            row = grid[key[0]]
            if not isinstance(row, Row):
                row = Row(
                    resource_name=self.resource_name
                    + "_layer_"
                    + str(key[2])
                    + "_row_"
                    + str(key[0]),
                    columns=self.columns,
                    children=list([None] * self.columns),
                )
                grid.children[key[0]] = row
            row[key[1]] = child
            child.key = str(key[2])

    def check_key_bounds(self, key: GridIndex3D) -> bool:
        """Check if the key is within the bounds of the grid."""

        if not isinstance(key, tuple) or len(key) != 3:
            raise ValueError("Key must be a 3-tuple.")
        return (
            not (key[2] < 0 or key[2] >= self.layers)
            and not (key[1] < 0 or key[1] >= self.rows)
            and not (key[0] < 0 or key[0] >= self.columns)
        )


class Slot(Container):
    """Data Model for a Slot. A slot is a container that can hold a single resource."""

    base_type: Literal[ContainerTypeEnum.slot] = Field(
        title="Container Base Type",
        description="The base type of the slot.",
        default=ContainerTypeEnum.slot,
        const=True,
    )
    children: list["ResourceDataModels"] = Field(
        title="Children",
        description="The children of the slot.",
        default_factory=list,
    )
    capacity: Literal[1] = Field(
        title="Capacity",
        description="The capacity of the slot.",
        default=1,
        const=1,
    )

    def get_child(self, key: Optional[int] = None) -> Optional["ResourceDataModels"]:
        """Get the child from the slot."""
        if key is not None and key != 0:
            return None
        return self.child

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        return len(self.children)

    @property
    def child(self) -> Optional["ResourceDataModels"]:
        """Get the child from the slot."""
        return self.children[0] if self.children else None

    def extract_children(self) -> dict[str, "ResourceDataModels"]:
        """Extract the children from the stack as a flat dict."""
        return {"0": self.child} if self.child else {}

    def populate_children(self, children: dict[str, "ResourceDataModels"]) -> None:
        """Populate the children of the stack."""
        self.children = [children["0"]] if "0" in children else []


class Stack(Container):
    """Data Model for a Stack. A stack is a container that can hold other resources in a single dimension and supports last-in, first-out (LIFO) access. For example, a stack of plates in a vertical magazine. Stacks are indexed by integers, with 0 being the bottom."""

    base_type: Literal[ContainerTypeEnum.stack] = Field(
        title="Container Base Type",
        description="The base type of the stack.",
        default=ContainerTypeEnum.stack,
        const=True,
    )
    children: list["ResourceDataModels"] = Field(
        title="Children",
        description="The children of the stack.",
        default_factory=list,
    )

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        return len(self.children)

    def get_child(self, key: int) -> Optional["ResourceDataModels"]:
        """Get a child from the container."""
        key = int(key)
        if key < 0 or key >= len(self.children):
            return None
        return self.children[key]

    def extract_children(self) -> dict[str, "ResourceDataModels"]:
        """Extract the children from the stack as a flat dict."""
        children_dict = {}
        for i in range(len(self.children)):
            children_dict[str(i)] = self.children[i]
        return children_dict

    def populate_children(self, children: dict[str, "ResourceDataModels"]) -> None:
        """Populate the children of the stack."""
        ordered_children = sorted(children.items(), key=lambda x: int(x[0]))
        self.children = [child[1] for child in ordered_children]


class Queue(Container):
    """Data Model for a Queue. A queue is a container that can hold other resources in a single dimension and supports first-in, first-out (FIFO) access. For example, a conveyer belt. Queues are indexed by integers, with 0 being the front."""

    base_type: Literal[ContainerTypeEnum.queue] = Field(
        title="Container Base Type",
        description="The base type of the queue.",
        default=ContainerTypeEnum.queue,
        const=True,
    )
    children: list["ResourceDataModels"] = Field(
        title="Children",
        description="The children of the queue.",
        default_factory=list,
    )

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        return len(self.children)

    def get_child(self, key: int) -> Optional["ResourceDataModels"]:
        """Get a child from the container."""
        key = int(key)
        if key < 0 or key >= len(self.children):
            return None
        return self.children[key]

    def extract_children(self) -> dict[str, "ResourceDataModels"]:
        """Extract the children from the stack as a flat dict."""
        children_dict = {}
        for i in range(len(self.children)):
            children_dict[str(i)] = self.children[i]
        return children_dict

    def populate_children(self, children: dict[str, "ResourceDataModels"]) -> None:
        """Populate the children of the queue."""
        ordered_children = sorted(children.items(), key=lambda x: int(x[0]))
        self.children = [child[1] for child in ordered_children]


class Pool(Container):
    """Data Model for a Pool. A pool is a container for holding consumables that can be mixed or collocated. For example, a single well in a microplate, or a reservoir. Pools are indexed by string key."""

    base_type: Literal[ContainerTypeEnum.pool] = Field(
        title="Container Base Type",
        description="The base type of the pool.",
        default=ContainerTypeEnum.pool,
        const=True,
    )
    children: dict[str, "ConsumableDataModels"] = Field(
        title="Children",
        description="The children of the pool.",
        default_factory=dict,
    )
    capacity: Optional[float] = Field(
        title="Capacity",
        description="The capacity of the pool as a whole.",
        default=None,
    )

    def get_child(self, key: str) -> Optional["ResourceDataModels"]:
        """Get a child from the container."""
        return self.children.get(key, None)

    @computed_field
    def quantity(self) -> int:
        """Calculate the quantity of assets in the container."""
        return sum(
            [
                child.quantity
                for child in self.children.values()
                if hasattr(child, "quantity")
            ]
        )

    def extract_children(self) -> dict[str, "ResourceDataModels"]:
        """Extract the children from the pool as a flat dictionary."""
        return self.children

    def populate_children(self, children: dict[str, "ResourceDataModels"]) -> None:
        """Populate the children of the pool."""
        self.children = children


RESOURCE_TYPE_MAP = {
    ResourceTypeEnum.resource: {
        "definition": ResourceDefinition,
        "model": Resource,
    },
    ResourceTypeEnum.asset: {
        "definition": AssetResourceDefinition,
        "model": Asset,
    },
    ResourceTypeEnum.consumable: {
        "definition": ConsumableResourceDefinition,
        "model": Consumable,
    },
    ConsumableTypeEnum.discrete_consumable: {
        "definition": DiscreteConsumableResourceDefinition,
        "model": DiscreteConsumable,
    },
    ConsumableTypeEnum.continuous_consumable: {
        "definition": ContinuousConsumableResourceDefinition,
        "model": ContinuousConsumable,
    },
    AssetTypeEnum.container: {
        "definition": ContainerResourceDefinition,
        "model": Container,
    },
    ContainerTypeEnum.stack: {
        "definition": StackResourceDefinition,
        "model": Stack,
    },
    ContainerTypeEnum.queue: {
        "definition": QueueResourceDefinition,
        "model": Queue,
    },
    ContainerTypeEnum.collection: {
        "definition": CollectionResourceDefinition,
        "model": Collection,
    },
    ContainerTypeEnum.row: {
        "definition": RowResourceDefinition,
        "model": Row,
    },
    ContainerTypeEnum.grid: {
        "definition": GridResourceDefinition,
        "model": Grid,
    },
    ContainerTypeEnum.voxel_grid: {
        "definition": VoxelGridResourceDefinition,
        "model": VoxelGrid,
    },
    ContainerTypeEnum.pool: {
        "definition": PoolResourceDefinition,
        "model": Pool,
    },
    ContainerTypeEnum.slot: {
        "definition": SlotResourceDefinition,
        "model": Slot,
    },
}

ResourceDataModels = Annotated[
    Union[
        Annotated[Resource, Tag("resource")],
        Annotated[Asset, Tag("asset")],
        Annotated[Consumable, Tag("consumable")],
        Annotated[DiscreteConsumable, Tag("discrete_consumable")],
        Annotated[ContinuousConsumable, Tag("continuous_consumable")],
        Annotated[Container, Tag("container")],
        Annotated[Collection, Tag("collection")],
        Annotated[Row, Tag("row")],
        Annotated[Grid, Tag("grid")],
        Annotated[VoxelGrid, Tag("voxel_grid")],
        Annotated[Stack, Tag("stack")],
        Annotated[Queue, Tag("queue")],
        Annotated[Pool, Tag("pool")],
        Annotated[Slot, Tag("slot")],
    ],
    Discriminator("base_type"),
]

ConsumableDataModels = Annotated[
    Union[
        Annotated[Consumable, Tag("consumable")],
        Annotated[DiscreteConsumable, Tag("discrete_consumable")],
        Annotated[ContinuousConsumable, Tag("continuous_consumable")],
    ],
    Discriminator("base_type"),
]

ContainerDataModels = Annotated[
    Union[
        Annotated[Container, Tag("container")],
        Annotated[Collection, Tag("collection")],
        Annotated[Row, Tag("row")],
        Annotated[Grid, Tag("grid")],
        Annotated[VoxelGrid, Tag("voxel_grid")],
        Annotated[Stack, Tag("stack")],
        Annotated[Queue, Tag("queue")],
        Annotated[Pool, Tag("pool")],
        Annotated[Slot, Tag("slot")],
    ],
    Discriminator("base_type"),
]



"""Utilities for the MADSci project."""

import json
import sys
from datetime import datetime, timezone
from pathlib import Path
from typing import Annotated, Any, Optional, Union, get_args, get_origin

from madsci.common.types.base_types import BaseModel, PathLike
from pydantic import ValidationError
from pydantic_core._pydantic_core import PydanticUndefined
from rich.console import Console

console = Console()


def utcnow() -> datetime:
    """Return the current UTC time."""

    return datetime.now(timezone.utc)


def localnow() -> datetime:
    """Return the current local time."""

    return datetime.now().astimezone()


def to_snake_case(name: str) -> str:
    """Convert a string to snake case.

    Handles conversion from camelCase and PascalCase to snake_case.
    """
    import re

    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", name)
    return name.lower().replace(" ", "_").replace("__", "_")


def search_for_file_pattern(
    pattern: str,
    start_dir: Optional[PathLike] = None,
    parents: bool = True,
    children: bool = True,
) -> list[str]:
    """
    Search up and down the file tree for a file(s) matching a pattern.

    Args:
        pattern: The pattern to search for. Standard glob patterns are supported.
        start_dir: The directory to start the search in. Defaults to the current directory.
        parents: Whether to search in parent directories.
        children: Whether to search in subdirectories.

    Returns:
        A list of paths to the files that match the pattern.
    """

    start_dir = Path.cwd() if not start_dir else Path(start_dir).expanduser()

    results = []
    if children:
        results.extend(Path("./").glob(str(Path("**") / pattern)))
    else:
        results.extend(Path.glob(str(Path("./") / pattern)))
        results.extend(Path("./").glob(pattern))
    if parents:
        for parent in start_dir.parents:
            results.extend(Path(parent).glob(pattern))
    return results


def save_model(path: PathLike, model: BaseModel, overwrite_check: bool = True) -> None:
    """Save a MADSci model to a YAML file, optionally with a check to overwrite if the file already exists."""
    try:
        model.model_validate(model)
    except ValidationError as e:
        raise ValueError(f"Validation error while saving model {model}: {e}") from e
    if (
        Path(path).exists()
        and overwrite_check
        and not prompt_yes_no(f"File already exists: {path}. Overwrite?", default="no")
    ):
        return
    model.to_yaml(path)


def prompt_yes_no(prompt: str, default: str = "no", quiet: bool = False) -> bool:
    """Prompt the user for a yes or no answer."""
    response = str(
        prompt_for_input(
            rf"{prompt} \[y/n]",
            default=default,
            required=False,
            quiet=quiet,
        ),
    ).lower()
    return response in ["y", "yes", "true"]


def prompt_for_input(
    prompt: str,
    default: Optional[str] = None,
    required: bool = False,
    quiet: bool = False,
) -> str:
    """Prompt the user for input."""
    if quiet or not sys.stdin.isatty():
        if default:
            return default
        if required:
            raise ValueError(
                "No input provided and no default value specified for required option.",
            )
        return None
    if not required:
        if default:
            response = console.input(f"{prompt} (optional, default: {default}): ")
        else:
            response = console.input(f"{prompt} (optional): ")
        if not response:
            response = default
    else:
        response = None
        while not response:
            if default:
                response = console.input(f"{prompt} (required, default: {default}): ")
                if not response:
                    response = default
            else:
                response = console.input(f"{prompt} (required): ")
    return response


def new_name_str(prefix: str = "") -> str:
    """Generate a new random name string, optionally with a prefix. Make a random combination of an adjective and a noun. Names are not guaranteed to be unique."""
    import random

    adjectives = [
        "happy",
        "clever",
        "bright",
        "swift",
        "calm",
        "bold",
        "eager",
        "fair",
        "kind",
        "proud",
        "brave",
        "wise",
        "quick",
        "sharp",
        "warm",
        "cool",
        "fresh",
        "keen",
        "agile",
        "gentle",
        "noble",
        "merry",
        "lively",
        "grand",
        "smart",
        "witty",
        "jolly",
        "mighty",
        "steady",
        "pure",
        "swift",
        "deft",
        "sage",
        "fleet",
        "spry",
        "bold",
    ]
    nouns = [
        "fox",
        "owl",
        "bear",
        "wolf",
        "hawk",
        "deer",
        "lion",
        "tiger",
        "eagle",
        "whale",
        "seal",
        "dove",
        "swan",
        "crow",
        "duck",
        "horse",
        "mouse",
        "cat",
        "lynx",
        "puma",
        "otter",
        "hare",
        "raven",
        "crane",
        "falcon",
        "badger",
        "marten",
        "stoat",
        "weasel",
        "vole",
        "rabbit",
        "squirrel",
        "raccoon",
        "beaver",
        "moose",
        "elk",
    ]

    name = f"{random.choice(adjectives)}_{random.choice(nouns)}"  # noqa: S311
    if prefix:
        name = f"{prefix}_{name}"
    return name


def string_to_bool(string: str) -> bool:
    """Convert a string to a boolean value."""
    from argparse import ArgumentTypeError

    if string.lower() in ("true", "t", "1", "yes", "y"):
        return True
    if string.lower() in ("false", "f", "0", "no", "n"):
        return False
    raise ArgumentTypeError(f"Invalid boolean value: {string}")


def prompt_from_list(
    prompt: str,
    options: list[str],
    default: Optional[str] = None,
    required: bool = False,
    quiet: bool = False,
) -> str:
    """Prompt the user for input from a list of options."""

    # *Print numbered list of options
    if not quiet:
        for i, option in enumerate(options, 1):
            console.print(f"[bold]{i}[/]. {option}")

    # *Allow selection by number or exact match
    def validate_response(response: str) -> Optional[str]:
        if response in options:
            return response
        try:
            idx = int(response)
            if 1 <= idx <= len(options):
                return options[idx - 1]
        except ValueError:
            pass
        return None

    while True:
        try:
            response = validate_response(
                prompt_for_input(
                    prompt,
                    default=default,
                    required=required,
                    quiet=quiet,
                ),
            )
        except ValueError:
            continue
        else:
            break
    return response


def prompt_from_pydantic_model(model: BaseModel, prompt: str, **kwargs: Any) -> str:
    """Prompt the user for input from a pydantic model.

    Args:
        model: The pydantic model to prompt for
        prompt: The prompt to display
        **kwargs: Pre-filled values to skip prompting for

    Returns:
        A dictionary of field values for the model
    """
    result = {}

    # Print header for the prompts
    console.print(f"\n[bold]{prompt}[/]")

    for field_name, field in model.__pydantic_fields__.items():
        # Skip if value provided in kwargs
        if field_name in kwargs:
            result[field_name] = kwargs[field_name]
            continue

        # Build field prompt
        field_prompt = f"{field.title or field_name}"

        # Add type hint
        type_hint = str(field.annotation).replace("typing.", "")
        field_prompt += f" ({type_hint})"

        # Add description if available
        if field.description:
            field_prompt += f"\n{field.description}"

        # Handle basic fields
        while True:
            try:
                response = prompt_for_input(
                    field_prompt,
                    default=field.default
                    if field.default != PydanticUndefined
                    else None,
                    required=field.is_required,
                )
                if isinstance(response, str):
                    response = json.loads(response)
                result[field_name] = response
            except json.JSONDecodeError as e:
                console.print(
                    f"[bold red]Invalid JSON input for field {field_name}: {e}[/]",
                )
                continue
            else:
                break

    return result


def relative_path(source: Path, target: Path, walk_up: bool = True) -> Path:
    """
    "Backport" of :meth:`pathlib.Path.relative_to` with ``walk_up=True``
    that's not available pre 3.12.

    Return the relative path to another path identified by the passed
    arguments.  If the operation is not possible (because this is not
    related to the other path), raise ValueError.

    The *walk_up* parameter controls whether `..` may be used to resolve
    the path.

    References:
        https://github.com/python/cpython/blob/8a2baedc4bcb606da937e4e066b4b3a18961cace/Lib/pathlib/_abc.py#L244-L270
    Credit: https://github.com/p2p-ld/numpydantic/blob/66fffc49f87bfaaa2f4d05bf1730c343b10c9cc6/src/numpydantic/serialization.py#L107
    """
    if not isinstance(source, Path):
        source = Path(source)
    target_parts = target.parts
    source_parts = source.parts
    anchor0, parts0 = target_parts[0], list(reversed(target_parts[1:]))
    anchor1, parts1 = source_parts[0], list(reversed(source_parts[1:]))
    if anchor0 != anchor1:
        raise ValueError(f"{target!r} and {source!r} have different anchors")
    while parts0 and parts1 and parts0[-1] == parts1[-1]:
        parts0.pop()
        parts1.pop()
    for part in parts1:
        if not part or part == ".":
            pass
        elif not walk_up:
            raise ValueError(f"{target!r} is not in the subpath of {source!r}")
        elif part == "..":
            raise ValueError(f"'..' segment in {source!r} cannot be walked")
        else:
            parts0.append("..")
    return Path(*reversed(parts0))


def threaded_task(func: callable) -> callable:
    """Mark a function as a threaded task, to be run without awaiting. Returns the thread object, so you _can_ await if needed."""

    import functools
    import threading

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> threading.Thread:
        thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        thread.start()
        return thread

    return wrapper


def threaded_daemon(func: callable) -> callable:
    """Mark a function as a threaded daemon, to be run without awaiting. Returns the thread object, so you _can_ await if needed, and stops when the calling thread terminates."""

    import functools
    import threading

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> threading.Thread:
        thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        thread.daemon = True
        thread.start()
        return thread

    return wrapper


def pretty_type_repr(type_hint: Any) -> str:
    """Returns a pretty string representation of a type hint, including subtypes."""
    type_name = getattr(type_hint, "__name__", None)
    if type_name is None:
        if get_origin(type_hint):
            type_name = get_origin(type_hint).__name__
        else:
            type_name = str(type_hint)
    if (
        "__args__" in dir(type_hint) and type_hint.__args__
    ):  # * If the type has subtype info
        type_name += "["
        for subtype in type_hint.__args__:
            type_name += pretty_type_repr(subtype)
            type_name += ", "
        type_name = type_name[:-2]
        type_name += "]"
    return type_name


@threaded_daemon
def repeat_on_interval(
    interval: float, func: callable, *args: Any, **kwargs: Any
) -> None:
    """Repeat a function on an interval."""
    import time

    while True:
        func(*args, **kwargs)
        time.sleep(interval)


def is_optional(type_hint: Any) -> bool:
    """Check if a type hint is Optional."""
    return get_origin(type_hint) is Union and type(None) in get_args(type_hint)


def is_annotated(type_hint: Any) -> bool:
    """Check if a type hint is an annotated type."""
    return get_origin(type_hint) is Annotated



"""Types for interacting with MADSci experiments and the Experiment Manager."""

from enum import Enum
from typing import ClassVar, Literal, Optional, Union

from bson.objectid import ObjectId
from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.base_types import BaseModel, datetime, new_ulid_str
from madsci.common.types.condition_types import Conditions
from madsci.common.types.event_types import EventClientConfig
from madsci.common.types.lab_types import ManagerDefinition, ManagerType
from pydantic import Field, field_validator


class ExperimentManagerDefinition(ManagerDefinition):
    """Definition for an Experiment Manager."""

    _definition_file_patterns: ClassVar[list] = [
        "*experiment_manager.yaml",
        "*experiment_manager.yml",
    ]
    _definition_cli_flags: ClassVar[list] = ["--experiment-manager", "--definition"]

    manager_type: Literal[ManagerType.EXPERIMENT_MANAGER] = Field(
        title="Manager Type",
        description="The type of the event manager",
        default=ManagerType.EXPERIMENT_MANAGER,
    )
    host: str = Field(
        title="Server Host",
        description="The host of the experiment manager.",
        default="127.0.0.1",
    )
    port: int = Field(
        title="Server Port",
        description="The port of the experiment manager.",
        default=8002,
    )
    db_url: str = Field(
        title="Database URL",
        description="The URL of the database for the experidict[str, Any]ment manager.",
        default="mongodb://localhost:27017",
    )
    lab_manager_url: Optional[str] = Field(
        title="Lab Manager URL", description="URL for the lab manager", default=None
    )
    workcell_manager_url: Optional[str] = Field(
        title="Workcell Manager URL",
        description="URL for the workcell manager",
        default=None,
    )
    resource_manager_url: Optional[str] = Field(
        title="Resource Manager URL",
        description="URL for the resource manager",
        default=None,
    )
    data_manager_url: Optional[str] = Field(
        title="Data Manager URL",
        description="URL for the data manager",
        default=None,
    )
    event_client_config: Optional[EventClientConfig] = Field(
        title="Event Client Configuration",
        description="The configuration for a MADSci event client.",
        default=None,
    )


class ExperimentDesign(BaseModel):
    """A design for a MADSci experiment."""

    experiment_name: str = Field(
        title="Experiment Name",
        description="The name of the experiment.",
    )
    experiment_description: Optional[str] = Field(
        title="Experiment Description",
        description="A description of the experiment.",
        default=None,
    )
    resource_conditions: list[Conditions] = Field(
        title="Resource Conditions",
        description="The starting layout of resources required for the experiment.",
        default_factory=list,
    )
    ownership_info: OwnershipInfo = Field(
        title="Ownership Info",
        description="Information about the users, campaigns, etc. that this design is owned by.",
        default_factory=OwnershipInfo,
    )
    event_client_config: Optional["EventClientConfig"] = Field(
        title="Event Client Configuration",
        description="The configuration for a MADSci event client.",
        default=None,
    )

    def new_experiment(
        self,
        run_name: Optional[str] = None,
        run_description: Optional[str] = None,
    ) -> "Experiment":
        """Create a new experiment from this design."""
        return Experiment.from_experiment_design(
            experiment_design=self, run_name=run_name, run_description=run_description
        )


class ExperimentRegistration(BaseModel):
    """Experiment Run Registration request body"""

    experiment_design: ExperimentDesign
    run_name: Optional[str] = None
    run_description: Optional[str] = None


class ExperimentStatus(str, Enum):
    """Current status of an experiment run."""

    IN_PROGRESS = "in_progress"
    """Experiment is currently running."""
    PAUSED = "paused"
    """Experiment is not currently running."""
    COMPLETED = "completed"
    """Experiment run has completed."""
    FAILED = "failed"
    """Experiment has failed."""
    CANCELLED = "cancelled"
    """Experiment has been cancelled."""
    UNKNOWN = "unknown"
    """Experiment status is unknown."""


class Experiment(BaseModel):
    """A MADSci experiment."""

    experiment_id: str = Field(
        title="Experiment ID",
        description="The ID of the experiment.",
        default_factory=new_ulid_str,
        alias="_id",
    )

    @field_validator("experiment_id", mode="before")
    @classmethod
    def object_id_to_str(cls, v: Union[str, ObjectId]) -> str:
        """Cast ObjectID to string."""
        if isinstance(v, ObjectId):
            return str(v)
        return v

    status: ExperimentStatus = Field(
        title="Experiment Status",
        description="The status of the experiment.",
        default=ExperimentStatus.IN_PROGRESS,
    )
    experiment_design: Optional[ExperimentDesign] = Field(
        title="Experiment Design",
        description="The design of the experiment.",
        default=None,
    )
    ownership_info: OwnershipInfo = Field(
        title="Ownership Info",
        description="Information about the ownership of the experiment.",
        default_factory=OwnershipInfo,
    )
    run_name: Optional[str] = Field(
        title="Run Name",
        description="A name for this specific experiment run.",
        default=None,
    )
    run_description: Optional[str] = Field(
        title="Run Description",
        description="A description of the experiment run.",
        default=None,
    )
    started_at: Optional[datetime] = Field(
        title="Started At",
        description="The time the experiment was started.",
        default=None,
    )
    ended_at: Optional[datetime] = Field(
        title="Ended At",
        description="The time the experiment was ended.",
        default=None,
    )

    @classmethod
    def from_experiment_design(
        cls,
        experiment_design: ExperimentDesign,
        run_name: Optional[str] = None,
        run_description: Optional[str] = None,
    ) -> "Experiment":
        """Create an experiment from an experiment design."""
        return cls(
            run_name=run_name,
            run_description=run_description,
            experiment_design=experiment_design,
            ownership_info=experiment_design.ownership_info.model_copy(),
        )


class ExperimentalCampaign(BaseModel):
    """A campaign consisting of one or more related experiments."""

    campaign_id: str = Field(
        title="Campaign ID",
        description="The ID of the campaign.",
        default_factory=new_ulid_str,
    )
    campaign_name: str = Field(
        title="Campaign Name",
        description="The name of the campaign.",
    )
    campaign_description: Optional[str] = Field(
        title="Campaign Description",
        description="A description of the campaign.",
        default=None,
    )
    experiment_ids: Optional[list[str]] = Field(
        title="Experiment IDs",
        description="The IDs of the experiments in the campaign. (Convenience field)",
        default_factory=None,
    )
    ownership_info: OwnershipInfo = Field(
        title="Ownership Info",
        description="Information about the ownership of the campaign.",
        default_factory=OwnershipInfo,
    )
    created_at: datetime = Field(
        title="Registered At",
        description="The time the campaign was registered.",
        default_factory=datetime.now,
    )
    ended_at: Optional[datetime] = Field(
        title="Ended At",
        description="The time the campaign was ended.",
        default=None,
    )



"""Types for MADSci Conditions."""

from enum import Enum
from typing import Annotated, Any, Literal, Optional, Union

from madsci.common.types.base_types import BaseModel
from madsci.common.types.resource_types import GridIndex, GridIndex2D, GridIndex3D
from pydantic import Discriminator
from sqlmodel.main import Field


class ConditionTypeEnum(str, Enum):
    """Types of conditional check for a step"""

    RESOURCE_PRESENT = "resource_present"
    NO_RESOURCE_PRESENT = "no_resource_present"
    RESOURCE_FIELD_CHECK = "resource_field_check"
    RESOURCE_CHILD_FIELD_CHECK = "resource_child_field_check"

    @classmethod
    def _missing_(cls, value: str) -> "ConditionTypeEnum":
        """Convert the value to lowercase"""
        value = value.lower()
        for member in cls:
            if member.lower() == value:
                return member
        raise ValueError(f"Invalid ConditionType: {value}")


class OperatorTypeEnum(str, Enum):
    """Comparison operators for value checks"""

    IS_GREATER_THAN = ("is_greater_than",)
    IS_LESS_THAN = ("is_less_than",)
    IS_EQUAL_TO = ("is_equal_to",)
    IS_GREQUAL_TO = "is_greater_than_or_equal_to"
    IS_LEQUAL_TO = "is_less_than_or_equal_to"


class Condition(BaseModel, extra="allow"):
    """A model for the conditions a step needs to be run"""

    condition_type: Optional[ConditionTypeEnum] = Field(
        title="Condition Type",
        description="The type of condition to check",
        default=None,
    )
    condition_name: str = Field(
        title="Condition Name",
        description="Name of the Condition",
        default="A Condition",
    )


class ResourceInLocationCondition(Condition):
    """A condition that checks if a resource is present"""

    condition_type: Literal[ConditionTypeEnum.RESOURCE_PRESENT] = Field(
        title="Condition Type",
        description="The type of condition to check",
        default=ConditionTypeEnum.RESOURCE_PRESENT,
    )
    location_id: Optional[str] = Field(
        title="Location",
        description="The ID of the location to check for a resource in",
        default=None,
    )

    location_name: Optional[str] = Field(
        title="Location",
        description="The name of the location to check for a resource in",
        default=None,
    )
    key: Union[str, int, GridIndex, GridIndex2D, GridIndex3D] = Field(
        title="Key",
        description="The key to check in the location's container resource",
        default=0,
    )
    resource_class: Optional[str] = Field(
        title="Resource Class",
        description="Check that the resource in this location is of a certain class",
        default=None,
    )


class NoResourceInLocationCondition(Condition):
    """A condition that checks if a resource is present"""

    condition_type: Literal[ConditionTypeEnum.NO_RESOURCE_PRESENT] = Field(
        title="Condition Type",
        description="The type of condition to check",
        default=ConditionTypeEnum.NO_RESOURCE_PRESENT,
    )

    location_id: Optional[str] = Field(
        title="Location",
        description="The ID of the location to check for a resource in",
        default=None,
    )

    location_name: str = Field(
        title="Location",
        description="The name of the location to check for a resource in",
    )
    key: Union[str, int, GridIndex, GridIndex2D, GridIndex3D] = Field(
        title="Key",
        description="The key to check in the location's container resource",
        default=0,
    )


class ResourceFieldCheckCondition(Condition):
    """A condition that checks if a resource is present"""

    condition_type: Literal[ConditionTypeEnum.RESOURCE_FIELD_CHECK] = Field(
        title="Condition Type",
        description="The type of condition to check",
        default=ConditionTypeEnum.RESOURCE_FIELD_CHECK,
    )
    resource_id: Optional[str] = Field(
        title="Resource ID",
        description="The id of the resource to check a quality of",
        default=None,
    )

    resource_name: Optional[str] = Field(
        title="Resource Name",
        description="The name of the resource to check a quality of",
        default=None,
    )
    field: str = Field(
        title="Field", description="The field to evaluate against the operator"
    )
    operator: OperatorTypeEnum = Field(
        title="Operator",
        description="The check (is_greater_than, is_less_than or is_equal_to etc.) to evaluate the field by",
    )
    target_value: Any = Field(
        title="Target Value", description="the target value for the field"
    )


class ResourceChildFieldCheckCondition(Condition):
    """A condition that checks if a resource is present"""

    condition_type: Literal[ConditionTypeEnum.RESOURCE_CHILD_FIELD_CHECK] = Field(
        title="Condition Type",
        description="The type of condition to check",
        default=ConditionTypeEnum.RESOURCE_CHILD_FIELD_CHECK,
    )
    resource_id: Optional[str] = Field(
        title="Resource ID",
        description="The id of the resource to check a quality of",
        default=None,
    )

    resource_name: Optional[str] = Field(
        title="Resource Name",
        description="The name of the resource to check a quality of",
        default=None,
    )
    field: str = Field(
        title="Field", description="The field to evaluate against the operator"
    )
    operator: OperatorTypeEnum = Field(
        title="Operator",
        description="The check (is_greater_than, is_less_than or is_equal_to etc.) to evaluate the field by",
    )
    key: Union[str, int, GridIndex, GridIndex2D, GridIndex3D] = Field(
        title="Key",
        description="The key to check in the container resource",
        default=0,
    )
    target_value: Any = Field(
        title="Target Value", description="the target value for the field"
    )


Conditions = Annotated[
    Union[
        ResourceInLocationCondition,
        NoResourceInLocationCondition,
        ResourceFieldCheckCondition,
        ResourceChildFieldCheckCondition,
    ],
    Discriminator(
        discriminator="condition_type",
    ),
]



"""Custom serializers for use in MADSci dataclasses."""

from typing import Any, Union


def dict_to_list(dct: Union[list[Any], dict[str, Any]]) -> list[Any]:
    """Converts a dictionary to a list of values.

    Example Usage:
        from pydantic import field_serializer

        serialize_nodes_to_list = field_serializer("nodes")(dict_to_list)
    """
    if isinstance(dct, list):
        return dct
    return list(dct.values())



"""Enums for MADSci Resource Types."""

from enum import Enum


class AssetTypeEnum(str, Enum):
    """Type for a MADSci Asset."""

    container = "container"
    asset = "asset"


class ConsumableTypeEnum(str, Enum):
    """Type for a MADSci Consumable."""

    consumable = "consumable"
    discrete_consumable = "discrete_consumable"
    continuous_consumable = "continuous_consumable"


class ContainerTypeEnum(str, Enum):
    """Type for a MADSci Container."""

    container = "container"
    slot = "slot"
    stack = "stack"
    queue = "queue"
    collection = "collection"
    row = "row"
    grid = "grid"
    voxel_grid = "voxel_grid"
    pool = "pool"


class ResourceTypeEnum(str, Enum):
    """Enum for all resource base types."""

    """Resource Base Types"""
    resource = "resource"

    """Asset Resource Base Types"""
    asset = "asset"
    container = "container"

    """Consumable Resource Base Types"""
    consumable = "consumable"
    discrete_consumable = "discrete_consumable"
    continuous_consumable = "continuous_consumable"

    """Container Resource Base Types"""
    slot = "slot"
    stack = "stack"
    queue = "queue"
    collection = "collection"
    row = "row"
    grid = "grid"
    voxel_grid = "voxel_grid"
    pool = "pool"



"""Types related to datapoint types"""

from datetime import datetime
from enum import Enum
from typing import Annotated, Any, Literal, Optional, Union

from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.base_types import BaseModel, PathLike, new_ulid_str
from madsci.common.types.event_types import EventClientConfig
from madsci.common.types.lab_types import ManagerDefinition, ManagerType
from pydantic import Field, Tag
from pydantic.types import Discriminator


class DataPointTypeEnum(str, Enum):
    """Enumeration for the types of data points.

    Attributes:
        FILE: Represents a data point that contains a file.
        DATA_VALUE: Represents a data point that contains a JSON serializable value.
    """

    FILE = "file"
    DATA_VALUE = "data_value"
    OBJECT_STORAGE = "object_storage"


class DataPoint(BaseModel, extra="allow"):
    """An object to contain and locate data created during experiments.

    Attributes:
        label: The label of this data point.
        step_id: The step that generated the data point.
        workflow_id: The workflow that generated the data point.
        experiment_id: The experiment that generated the data point.
        campaign_id: The campaign of the data point.
        data_type: The type of the data point, inherited from class.
        datapoint_id: The specific ID for this data point.
        data_timestamp: The time the data point was created.
    """

    label: str
    """Label of this data point"""
    ownership_info: Optional[OwnershipInfo] = Field(default_factory=OwnershipInfo)
    """Information about the ownership of the data point"""
    data_type: DataPointTypeEnum
    """type of the datapoint, inherited from class"""
    datapoint_id: str = Field(default_factory=new_ulid_str)
    """specific id for this data point"""
    data_timestamp: datetime = Field(default_factory=datetime.now)
    """time datapoint was created"""

    @classmethod
    def discriminate(cls, datapoint: "DataPointDataModels") -> "DataPointDataModels":
        """Return the correct data point type based on the data_type attribute.

        Args:
            datapoint: The data point instance or dictionary to discriminate.

        Returns:
            The appropriate DataPoint subclass instance.
        """
        if isinstance(datapoint, dict):
            datapoint_type = datapoint["data_type"]
        else:
            datapoint_type = datapoint.data_type
        return DataPointTypeMap[datapoint_type].model_validate(datapoint)


class FileDataPoint(DataPoint):
    """A data point containing a file.

    Attributes:
        data_type: The type of the data point, in this case a file.
        path: The path to the file.
    """

    data_type: Literal[DataPointTypeEnum.FILE] = DataPointTypeEnum.FILE
    """The type of the data point, in this case a file"""
    path: PathLike
    """Path to the file"""


class ValueDataPoint(DataPoint):
    """A data point corresponding to a single JSON serializable value.

    Attributes:
        data_type: The type of the data point, in this case a value.
        value: The value of the data point.
    """

    data_type: Literal[DataPointTypeEnum.DATA_VALUE] = DataPointTypeEnum.DATA_VALUE
    """The type of the data point, in this case a value"""
    value: Any
    """Value of the data point"""


class ObjectStorageDataPoint(DataPoint):
    """A data point that references an object in S3-compatible storage (MinIO/S3).

    This data point stores essential information about an object in S3-compatible
    storage without storing access credentials.

    Attributes:
        url: The accessible URL for the object (can be used in frontend).
        storage_endpoint: The endpoint of the storage service (e.g., 'minio.example.com:9000').
        bucket_name: The name of the bucket containing the object.
        object_name: The path/key of the object within the bucket.
        content_type: The MIME type of the stored object.
        size_bytes: The size of the object in bytes.
        etag: The entity tag (typically MD5) of the object.
        custom_metadata: Additional user-defined metadata for the object.
    """

    url: Optional[str] = Field(
        default=None, description="Accessible URL for the object (for frontend use)"
    )
    data_type: Literal[DataPointTypeEnum.OBJECT_STORAGE] = (
        DataPointTypeEnum.OBJECT_STORAGE
    )
    """The type of the data point, in this case an object storage"""
    storage_endpoint: str = Field(
        ..., description="S3 API endpoint (e.g., 'localhost:9000')"
    )
    public_endpoint: Optional[str] = Field(
        default=None,
        description="Public endpoint for accessing objects (e.g., 'localhost:9001')",
    )
    path: PathLike
    """Path to the file"""
    bucket_name: str = Field(
        default=None, description="Name of the bucket containing the object"
    )
    object_name: str = Field(
        default=None, description="Path/key of the object within the bucket"
    )
    content_type: Optional[str] = Field(
        None, description="MIME type of the stored object"
    )
    size_bytes: Optional[int] = Field(None, description="Size of the object in bytes")
    etag: Optional[str] = Field(
        None, description="Entity tag (typically MD5) of the object"
    )
    custom_metadata: dict[str, str] = Field(
        default_factory=dict, description="User-defined metadata for the object"
    )


DataPointDataModels = Annotated[
    Union[
        Annotated[FileDataPoint, Tag(DataPointTypeEnum.FILE)],
        Annotated[ValueDataPoint, Tag(DataPointTypeEnum.DATA_VALUE)],
        Annotated[ObjectStorageDataPoint, Tag(DataPointTypeEnum.OBJECT_STORAGE)],
    ],
    Discriminator("data_type"),
]

DataPointTypeMap = {
    DataPointTypeEnum.FILE: FileDataPoint,
    DataPointTypeEnum.DATA_VALUE: ValueDataPoint,
    DataPointTypeEnum.OBJECT_STORAGE: ObjectStorageDataPoint,
}


class ObjectStorageDefinition(BaseModel):
    """Configuration for S3-compatible object storage."""

    endpoint: str
    """Endpoint for S3-compatible storage (e.g., 'minio.example.com:9000')"""
    access_key: str
    """Access key for authentication"""
    secret_key: str
    """Secret key for authentication"""
    secure: bool = False
    """Whether to use HTTPS (True) or HTTP (False)"""
    default_bucket: str = "madsci-data"
    """Default bucket to use for storing data"""
    region: Optional[str] = None
    """Optional for AWS S3/other providers"""


class DataManagerDefinition(ManagerDefinition):
    """Definition for a Squid Data Manager.

    Attributes:
        manager_type: The type of the event manager.
        host: The hostname or IP address of the Data Manager server.
        port: The port number of the Data Manager server.
        db_url: The URL of the database used by the Data Manager.
        event_client_config: The configuration for a MADSci event client.
    """

    manager_type: Literal[ManagerType.DATA_MANAGER] = Field(
        title="Manager Type",
        description="The type of the event manager",
        default=ManagerType.DATA_MANAGER,
    )
    host: str = Field(
        default="127.0.0.1",
        title="Server Host",
        description="The hostname or IP address of the Data Manager server.",
    )
    port: int = Field(
        default=8004,
        title="Server Port",
        description="The port number of the Data Manager server.",
    )
    db_url: str = Field(
        default="mongodb://localhost:27017",
        title="Database URL",
        description="The URL of the database used by the Data Manager.",
    )
    file_storage_path: PathLike = Field(
        title="File Storage Path",
        description="The path where files are stored on the server.",
        default="~/.madsci/datapoints",
    )
    event_client_config: Optional[EventClientConfig] = Field(
        title="Event Client Configuration",
        description="The configuration for a MADSci event client.",
        default=None,
    )
    minio_client_config: Optional[ObjectStorageDefinition] = Field(
        title="MinIO Client Configuration",
        description="Configuration for MinIO client for object storage.",
        default=None,
    )



"""Types for MADSci Squid Lab configuration."""

from enum import Enum
from pathlib import Path
from typing import ClassVar, Optional

from madsci.common.types.base_types import BaseModel, PathLike, new_ulid_str
from madsci.common.validators import (
    ulid_validator,
)
from pydantic import ConfigDict
from pydantic.functional_validators import field_validator
from pydantic.networks import AnyUrl
from sqlmodel.main import Field


class LabUrls(BaseModel):
    """urls for the lab manager"""

    workcell_manager: Optional[AnyUrl] = Field(
        title="Workcell Manager URL",
        description="The Workcell Manager Url",
        default=None,
    )

    resource_manager: Optional[AnyUrl] = Field(
        title="Resource Manager URL",
        description="The Resource Manager URL for this lab.",
        default=None,
    )

    data_manager: Optional[AnyUrl] = Field(
        title="Data Manager URL",
        description="The Data Manager URL for this lab.",
        default=None,
    )

    experiment_manager: Optional[AnyUrl] = Field(
        title="Data Manager URL",
        description="The Data Manager URL for this lab.",
        default=None,
    )


class LabDefinition(BaseModel):
    """Definition for a MADSci Lab."""

    model_config = ConfigDict(extra="allow")
    _definition_file_patterns: ClassVar[list[str]] = ["*.lab.yaml"]

    name: str = Field(title="Name", description="The name of the lab.")
    lab_id: str = Field(
        title="Lab ID",
        description="The ID of the lab.",
        default_factory=new_ulid_str,
    )
    description: Optional[str] = Field(
        default=None,
        title="Description",
        description="A description of the lab.",
    )
    host: str = Field(
        default="127.0.0.1",
        title="Server Host",
        description="The hostname or IP address of the Squid Lab Server.",
    )
    port: int = Field(
        default=8000,
        title="Server Port",
        description="The port number of the Squid Lab Server.",
    )
    commands: dict[str, str] = Field(
        default_factory=dict,
        title="Commands",
        description="Commands for operating the lab.",
    )
    static_files_path: Optional[PathLike] = Field(
        default=Path("~") / "MADSci" / "src" / "madsci_squid" / "ui" / "dist",
        title="Static Files Path",
        description="Path to the static files for the lab manager. Set to None to disable.",
    )
    managers: "LabUrls" = Field(
        default_factory=LabUrls,
        title="Manager URLs",
        description="URLs for Managers used by the lab.",
    )

    @field_validator("commands")
    @classmethod
    def validate_commands(cls, v: dict[str, str]) -> dict[str, str]:
        """Validate the command names."""
        if v:
            for command in v:
                if not str(command).replace("_", "").isalnum():
                    raise ValueError(
                        f"Command '{command}' must consist only of alphanumeric characters or underscores."
                    )
        return v

    is_ulid = field_validator("lab_id")(ulid_validator)

    def resolve_managers(self, path_origin: Optional[PathLike] = None) -> None:
        """Resolve the manager definition links."""
        if path_origin is not None:
            if self._definition_path:
                path_origin = Path(self._definition_path)
            else:
                path_origin = Path("./")
        for manager in self.managers.values():
            manager.resolve(path_origin=path_origin)


class ManagerDefinition(BaseModel):
    """Definition for a Squid Manager."""

    model_config = ConfigDict(extra="allow")
    _definition_file_patterns: ClassVar[list[str]] = ["*.manager.yaml"]

    name: str = Field(
        title="Manager Name",
        description="The name of this manager instance.",
    )
    manager_id: Optional[str] = Field(
        title="Manager ID",
        description="The ID of the manager.",
        default_factory=new_ulid_str,
    )
    description: Optional[str] = Field(
        default=None,
        title="Description",
        description="A description of the manager.",
    )
    manager_type: "ManagerType" = Field(
        title="Manager Type",
        description="The type of the manager, used by other components or managers to find matching managers.",
    )
    url: Optional[AnyUrl] = Field(
        default=None,
        title="Manager URL",
        description="The URL of the manager's API.",
    )


class ManagerType(str, Enum):
    """Types of Squid Managers."""

    WORKCELL_MANAGER = "workcell_manager"
    RESOURCE_MANAGER = "resource_manager"
    EVENT_MANAGER = "event_manager"
    AUTH_MANAGER = "auth_manager"
    NOTIFICATION_MANAGER = "notification_manager"
    DATA_MANAGER = "data_manager"
    TRANSFER_MANAGER = "transfer_manager"
    DASHBOARD_MANAGER = "dashboard_manager"
    EXPERIMENT_MANAGER = "experiment_manager"

    @classmethod
    def _missing_(cls, value: str) -> "ManagerType":
        value = value.lower()
        for member in cls:
            if member.lower() == value:
                return member
        raise ValueError(f"Invalid ManagerTypes: {value}")



"""Types for MADSci Actions."""

from datetime import datetime
from enum import Enum
from typing import Any, Literal, Optional, Union

from madsci.common.types.base_types import BaseModel, Error, PathLike, new_ulid_str
from madsci.common.types.datapoint_types import DataPoint
from madsci.common.utils import localnow
from pydantic import Field
from pydantic.functional_validators import field_validator, model_validator


class ActionStatus(str, Enum):
    """Status for a step of a workflow"""

    NOT_STARTED = "not_started"
    NOT_READY = "not_ready"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    CANCELLED = "cancelled"
    PAUSED = "paused"
    UNKNOWN = "unknown"

    @property
    def is_terminal(self) -> bool:
        """Check if the status is terminal"""
        return self in [
            ActionStatus.SUCCEEDED,
            ActionStatus.FAILED,
            ActionStatus.CANCELLED,
            ActionStatus.NOT_READY,
        ]


class ActionRequest(BaseModel):
    """Request to perform an action on a node"""

    action_id: str = Field(
        title="Action ID",
        description="The ID of the action.",
        default_factory=new_ulid_str,
    )
    action_name: str = Field(
        title="Action Name",
        description="The name of the action to perform.",
    )
    """Name of the action to perform"""
    args: Optional[dict[str, Any]] = Field(
        title="Action Arguments",
        description="Arguments for the action.",
        default_factory=dict,
    )
    """Arguments for the action"""
    files: dict[str, PathLike] = Field(
        title="Action Files",
        description="Files sent along with the action.",
        default_factory=dict,
    )
    """Files sent along with the action"""

    def failed(
        self,
        errors: Union[Error, list[Error], str] = [],
        data: dict[str, Any] = {},
        files: dict[str, PathLike] = {},
    ) -> "ActionFailed":
        """Create an ActionFailed response"""
        # * Convert errors to a list of errors if they are a single error or a string
        if isinstance(errors, str):
            errors = [Error(message=errors)]
        elif isinstance(errors, Error):
            errors = [errors]
        return ActionFailed(
            action_id=self.action_id,
            errors=errors,
            data=data,
            files=files,
        )

    def succeeded(
        self,
        data: dict[str, Any] = {},
        files: dict[str, PathLike] = {},
        errors: Union[Error, list[Error], str] = [],
    ) -> "ActionSucceeded":
        """Create an ActionSucceeded response"""
        return ActionSucceeded(
            action_id=self.action_id,
            errors=errors,
            data=data,
            files=files,
        )

    def running(
        self,
        data: dict[str, Any] = {},
        files: dict[str, PathLike] = {},
        errors: Union[Error, list[Error], str] = [],
    ) -> "ActionRunning":
        """Create an ActionRunning response"""
        return ActionRunning(
            action_id=self.action_id,
            errors=errors,
            data=data,
            files=files,
        )

    def not_ready(
        self,
        errors: Union[Error, list[Error], str] = [],
        data: dict[str, Any] = {},
        files: dict[str, PathLike] = {},
    ) -> "ActionNotReady":
        """Create an ActionNotReady response"""
        return ActionNotReady(
            action_id=self.action_id,
            errors=errors,
            data=data,
            files=files,
        )

    def cancelled(
        self,
        errors: Union[Error, list[Error], str] = [],
        data: dict[str, Any] = {},
        files: dict[str, PathLike] = {},
    ) -> "ActionCancelled":
        """Create an ActionCancelled response"""
        return ActionCancelled(
            action_id=self.action_id,
            errors=errors,
            data=data,
            files=files,
        )

    def paused(
        self,
        errors: Union[Error, list[Error], str] = [],
        data: dict[str, Any] = {},
        files: dict[str, PathLike] = {},
    ) -> "ActionResult":
        """Create an ActionResult response"""
        return ActionResult(
            action_id=self.action_id,
            status=ActionStatus.PAUSED,
            errors=errors,
            data=data,
            files=files,
        )

    def not_started(
        self,
        errors: Union[Error, list[Error], str] = [],
        data: dict[str, Any] = {},
        files: dict[str, PathLike] = {},
    ) -> "ActionResult":
        """Create an ActionResult response"""
        return ActionResult(
            action_id=self.action_id,
            status=ActionStatus.NOT_STARTED,
            errors=errors,
            data=data,
            files=files,
        )

    def unknown(
        self,
        errors: Union[Error, list[Error], str] = [],
        data: dict[str, Any] = {},
        files: dict[str, PathLike] = {},
    ) -> "ActionResult":
        """Create an ActionResult response"""
        return ActionResult(
            action_id=self.action_id,
            status=ActionStatus.UNKNOWN,
            errors=errors,
            data=data,
            files=files,
        )


class ActionResult(BaseModel):
    """Result of an action."""

    action_id: str = Field(
        title="Action ID",
        description="The ID of the action.",
        default_factory=new_ulid_str,
    )
    status: ActionStatus = Field(
        title="Step Status",
        description="The status of the step.",
    )
    errors: list[Error] = Field(
        title="Step Error",
        description="An error message(s) if the step failed.",
        default_factory=list,
    )
    data: dict[str, Any] = Field(
        title="Step Data",
        description="The data generated by the step.",
        default_factory=dict,
    )
    files: dict[str, PathLike] = Field(
        title="Step Files",
        description="A dictionary of files produced by the step.",
        default_factory=dict,
    )
    datapoints: dict[str, DataPoint] = Field(
        title="Data Points",
        description="A dictionary of datapoints sent to the data manager by the step.",
        default_factory=dict,
    )
    history_created_at: Optional[datetime] = Field(
        title="History Created At",
        description="The time the history was updated.",
        default_factory=localnow,
    )

    @field_validator("errors", mode="before")
    @classmethod
    def ensure_list_of_errors(cls, v: Any) -> Any:
        """Ensure that errors is a list of MADSci Errors"""
        if isinstance(v, str):
            return [Error(message=v)]
        if isinstance(v, Error):
            return [v]
        if isinstance(v, Exception):
            return [Error.from_exception(v)]
        if isinstance(v, list):
            for i, item in enumerate(v):
                if isinstance(item, str):
                    v[i] = Error(message=item)
                elif isinstance(item, Exception):
                    v[i] = Error.from_exception(item)
        return v


class ActionSucceeded(ActionResult):
    """Response from an action that succeeded."""

    status: Literal[ActionStatus.SUCCEEDED] = ActionStatus.SUCCEEDED


class ActionFailed(ActionResult):
    """Response from an action that failed."""

    status: Literal[ActionStatus.FAILED] = ActionStatus.FAILED


class ActionCancelled(ActionResult):
    """Response from an action that was cancelled."""

    status: Literal[ActionStatus.CANCELLED] = ActionStatus.CANCELLED


class ActionRunning(ActionResult):
    """Response from an action that is running."""

    status: Literal[ActionStatus.RUNNING] = ActionStatus.RUNNING


class ActionNotReady(ActionResult):
    """Response from an action that is not ready to be run."""

    status: Literal[ActionStatus.NOT_READY] = ActionStatus.NOT_READY


class ActionPaused(ActionResult):
    """Response from an action that is paused."""

    status: Literal[ActionStatus.PAUSED] = ActionStatus.PAUSED


class ActionNotStarted(ActionResult):
    """Response from an action that has not started."""

    status: Literal[ActionStatus.NOT_STARTED] = ActionStatus.NOT_STARTED


class ActionUnknown(ActionResult):
    """Response from an action that has an unknown status."""

    status: Literal[ActionStatus.UNKNOWN] = ActionStatus.UNKNOWN


class ActionDefinition(BaseModel):
    """Definition of an action."""

    name: str = Field(
        title="Action Name",
        description="The name of the action.",
    )
    description: str = Field(
        title="Action Description",
        description="A description of the action.",
    )

    @field_validator("description", mode="before")
    @classmethod
    def none_to_empty_str(cls, v: Any) -> str:
        """Convert None to empty string"""
        if v is None:
            return ""
        return v

    args: Union[
        dict[str, "ArgumentDefinition"],
        list["ArgumentDefinition"],
    ] = Field(
        title="Action Arguments",
        description="The arguments of the action.",
        default_factory=dict,
    )
    locations: Union[
        dict[str, "LocationArgumentDefinition"], list["LocationArgumentDefinition"]
    ] = Field(
        title="Action Location Arguments",
        description="The location arguments of the action.",
        default_factory=dict,
    )
    files: Union[
        dict[str, "FileArgumentDefinition"], list["FileArgumentDefinition"]
    ] = Field(
        title="Action File Arguments",
        description="The file arguments of the action.",
        default_factory=dict,
    )
    results: Union[
        dict[str, "ActionResultDefinition"],
        list["ActionResultDefinition"],
    ] = Field(
        title="Action Results",
        description="The results of the action.",
        default_factory=dict,
    )
    blocking: bool = Field(
        title="Blocking",
        description="Whether the action is blocking.",
        default=False,
    )
    asynchronous: bool = Field(
        title="Asynchronous",
        description="Whether the action is asynchronous, and will return a 'running' status immediately rather than waiting for the action to complete before returning. This should be used for long-running actions (e.g. actions that take more than a few seconds to complete).",
        default=True,
    )

    @field_validator("args", mode="after")
    @classmethod
    def ensure_args_are_dict(cls, v: Any) -> Any:
        """Ensure that the args are a dictionary"""
        if isinstance(v, list):
            return {arg.name: arg for arg in v}
        return v

    @field_validator("files", mode="after")
    @classmethod
    def ensure_files_are_dict(cls, v: Any) -> Any:
        """Ensure that the files are a dictionary"""
        if isinstance(v, list):
            return {file.name: file for file in v}
        return v

    @field_validator("locations", mode="after")
    @classmethod
    def ensure_locations_are_dict(cls, v: Any) -> Any:
        """Ensure that the locations are a dictionary"""
        if isinstance(v, list):
            return {location.name: location for location in v}
        return v

    @field_validator("results", mode="after")
    @classmethod
    def ensure_results_are_dict(cls, v: Any) -> Any:
        """Ensure that the results are a dictionary"""
        if isinstance(v, list):
            return {result.result_label: result for result in v}
        return v

    @model_validator(mode="after")
    @classmethod
    def ensure_name_uniqueness(cls, v: Any) -> Any:
        """Ensure that the names of the arguments and files are unique"""
        names = set()
        for arg in v.args.values():
            if arg.name in names:
                raise ValueError(f"Action name '{arg.name}' is not unique")
            names.add(arg.name)
        for file in v.files.values():
            if file.name in names:
                raise ValueError(f"File name '{file.name}' is not unique")
            names.add(file.name)
        for location in v.locations.values():
            if location.name in names:
                raise ValueError(f"Location name '{location.name}' is not unique")
            names.add(location.name)
        return v


class ArgumentDefinition(BaseModel):
    """Defines an argument for a node action"""

    name: str = Field(
        title="Argument Name",
        description="The name of the argument.",
    )
    description: str = Field(
        title="Argument Description",
        description="A description of the argument.",
    )
    argument_type: str = Field(
        title="Argument Type", description="Any type information about the argument"
    )
    required: bool = Field(
        title="Argument Required",
        description="Whether the argument is required.",
    )
    default: Optional[Any] = Field(
        title="Argument Default",
        description="The default value of the argument.",
        default=None,
    )


class LocationArgumentDefinition(ArgumentDefinition):
    """Location Argument Definition for use in NodeInfo"""

    argument_type: Literal["location"] = Field(
        title="Location Argument Type",
        description="The type of the location argument.",
        default="location",
    )


class FileArgumentDefinition(ArgumentDefinition):
    """Defines a file for a node action"""

    argument_type: Literal["file"] = Field(
        title="File Argument Type",
        description="The type of the file argument.",
        default="file",
    )


class ActionResultDefinition(BaseModel):
    """Defines a result for a node action"""

    result_label: str = Field(
        title="Result Label",
        description="The label of the result.",
    )
    description: str = Field(
        title="Result Description",
        description="A description of the result.",
        default=None,
    )
    result_type: str = Field(
        title="Result Type",
        description="The type of the result.",
    )


class FileActionResultDefinition(ActionResultDefinition):
    """Defines a file result for a node action"""

    result_type: Literal["file"] = Field(
        title="Result Type",
        description="The type of the result.",
        default="file",
    )


class JSONActionResultDefinition(ActionResultDefinition):
    """Defines a JSON result for a node action"""

    result_type: Literal["json"] = Field(
        title="Result Type",
        description="The type of the result.",
        default="json",
    )



"MADSci Node Types."

from datetime import datetime
from enum import Enum
from typing import Any, ClassVar, Optional

from madsci.common.types.action_types import ActionDefinition
from madsci.common.types.admin_command_types import AdminCommands
from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.base_types import BaseModel, Error, new_ulid_str
from madsci.common.types.datapoint_types import ObjectStorageDefinition
from madsci.common.types.event_types import EventClientConfig
from madsci.common.validators import ulid_validator
from pydantic import (
    SerializationInfo,
    SerializerFunctionWrapHandler,
    field_serializer,
    model_serializer,
)
from pydantic.config import ConfigDict
from pydantic.fields import computed_field
from pydantic.functional_validators import field_validator
from pydantic.networks import AnyUrl
from pydantic_extra_types.semantic_version import SemanticVersion
from semver import Version
from sqlmodel.main import Field


class NodeType(str, Enum):
    """The type of a MADSci node."""

    DEVICE = "device"
    COMPUTE = "compute"
    RESOURCE_MANAGER = "resource_manager"
    EVENT_MANAGER = "event_manager"
    WORKCELL_MANAGER = "workcell_manager"
    DATA_MANAGER = "data_manager"
    TRANSFER_MANAGER = "transfer_manager"


class NodeConfig(BaseModel):
    """Basic Configuration for a MADSci Node."""

    model_config = ConfigDict(extra="allow")

    status_update_interval: Optional[float] = Field(
        title="Status Update Interval",
        description="The interval in seconds at which the node should update its status.",
        default=2.0,
    )
    state_update_interval: Optional[float] = Field(
        title="State Update Interval",
        description="The interval in seconds at which the node should update its state.",
        default=2.0,
    )
    event_client_config: Optional[EventClientConfig] = Field(
        title="Event Client Configuration",
        description="The configuration for a MADSci event client.",
        default=None,
    )
    resource_server_url: Optional[AnyUrl] = Field(
        title="Resource Client URL",
        description="The URL of the resource server for this node to use.",
        default=None,
    )
    minio_client_config: Optional[ObjectStorageDefinition] = Field(
        title="Object Storage Configuration",
        description="Configuration for S3-compatible object storage using MinIO.",
        default=None,
    )


class RestNodeConfig(NodeConfig):
    """Default Configuration for a MADSci Node that communicates over REST."""

    host: str = Field(
        title="Host",
        description="The host of the REST API.",
        default="127.0.0.1",
    )
    port: int = Field(
        title="Port",
        description="The port of the REST API.",
        default=2000,
    )
    protocol: str = Field(
        title="Protocol",
        description="The protocol of the REST API, either 'http' or 'https'.",
        default="http",
    )


class NodeClientCapabilities(BaseModel):
    """Capabilities of a MADSci Node Client. Default values are None, meaning the capability is not explicitly set. If a capability is set to False, it is explicitly not supported."""

    get_info: Optional[bool] = Field(
        default=None,
        title="Node Info",
        description="Whether the node supports querying its info.",
    )
    get_state: Optional[bool] = Field(
        default=None,
        title="Node State",
        description="Whether the node supports querying its state.",
    )
    get_status: Optional[bool] = Field(
        default=None,
        title="Node Status",
        description="Whether the node supports querying its status.",
    )
    send_action: Optional[bool] = Field(
        default=None,
        title="Node Send Action",
        description="Whether the node supports sending actions.",
    )
    get_action_result: Optional[bool] = Field(
        default=None,
        title="Node Get Action",
        description="Whether the node supports querying the status of an action.",
    )
    get_action_history: Optional[bool] = Field(
        default=None,
        title="Node Get Actions",
        description="Whether the node supports querying the history of actions.",
    )
    action_files: Optional[bool] = Field(
        default=None,
        title="Node Action Files",
        description="Whether the node supports sending action files.",
    )
    send_admin_commands: Optional[bool] = Field(
        default=None,
        title="Node Send Admin Commands",
        description="Whether the node supports sending admin commands.",
    )
    set_config: Optional[bool] = Field(
        default=None,
        title="Node Set Config",
        description="Whether the node supports setting configuration.",
    )
    get_resources: Optional[bool] = Field(
        default=None,
        title="Node Get Resources",
        description="Whether the node supports querying its resources.",
    )
    get_log: Optional[bool] = Field(
        default=None,
        title="Node Get Log",
        description="Whether the node supports querying its log.",
    )

    @model_serializer(mode="wrap")
    def exclude_unset_by_default(
        self, nxt: SerializerFunctionWrapHandler, info: SerializationInfo
    ) -> dict[str, Any]:
        """Exclude unset fields by default."""
        serialized = nxt(self, info)
        return {k: v for k, v in serialized.items() if v is not None}


class NodeCapabilities(NodeClientCapabilities):
    """Capabilities of a MADSci Node."""

    events: Optional[bool] = Field(
        default=None,
        title="Node Events",
        description="Whether the node supports raising MADSci events.",
    )
    resources: Optional[bool] = Field(
        default=None,
        title="Node Resources",
        description="Whether the node supports MADSci-compatible resource management.",
    )
    admin_commands: set[AdminCommands] = Field(
        default=set(),
        title="Node Admin Commands",
        description="Which admin commands the node supports, if any.",
    )

    @field_serializer("admin_commands")
    def order_admin_commands(
        self, admin_commands: set[AdminCommands]
    ) -> list[AdminCommands]:
        """Ensure sorted admin commands."""
        return sorted(admin_commands)


class NodeDefinition(BaseModel):
    """Definition of a MADSci Node, a unique instance of a MADSci Node Module."""

    _definition_file_patterns: ClassVar[list[str]] = ["*.node.yaml"]

    node_name: str = Field(title="Node Name", description="The name of the node.")
    node_id: str = Field(
        title="Node ID", description="The ID of the node.", default_factory=new_ulid_str
    )
    node_url: Optional[AnyUrl] = Field(
        title="Node URL",
        description="The URL used to communicate with the node.",
        default=None,
    )
    node_description: Optional[str] = Field(
        title="Description",
        description="A description of the node.",
        default=None,
    )
    node_type: Optional[NodeType] = Field(
        title="Node Type",
        description="The type of thing this node provides an interface for.",
        default=None,
    )
    module_name: str = Field(
        title="Node Module Name",
        description="The name of the node module.",
    )
    module_version: SemanticVersion = Field(
        default=Version.parse("0.0.1"),
        title="Module Version",
        description="The version of the node module.",
    )
    capabilities: Optional["NodeCapabilities"] = Field(
        default=None,
        title="Node Capabilities",
        description="Explicitly override the capabilities of the node.",
    )
    commands: dict[str, str] = Field(
        title="Node Commands",
        description="The commands that the node supports. These can be used to stop, start, or otherwise administrate the node(s).",
        default_factory=dict,
    )
    is_template: bool = Field(
        default=False,
        title="Is Template",
        description="Whether this is a template definition.",
    )
    config_defaults: dict[str, Any] = Field(
        default_factory=dict,
        title="Node Configuration Defaults",
        description="Default values to use for configuration parameters for this particular node.",
    )

    is_ulid = field_validator("node_id")(ulid_validator)

    @model_serializer(mode="wrap")
    def exclude_node_id_on_templates(
        self, nxt: SerializerFunctionWrapHandler, info: SerializationInfo
    ) -> dict[str, Any]:
        """Exclude node_id on templates."""
        serialized = nxt(self, info)
        if self.is_template:
            serialized.pop("node_id")
        return serialized


class Node(BaseModel, arbitrary_types_allowed=True):
    """A runtime representation of a MADSci Node used in a Workcell."""

    node_url: AnyUrl = Field(
        title="Node URL",
        description="The URL used to communicate with the node.",
    )
    status: Optional["NodeStatus"] = Field(
        default=None,
        title="Node Status",
        description="The status of the node. Set to None if the node does not support status reporting or the status is unknown (e.g. if it hasn't reported/responded to status requests).",
    )
    info: Optional["NodeInfo"] = Field(
        default=None,
        title="Node Info",
        description="Information about the node, provided by the node itself.",
    )
    state: Optional[dict[str, Any]] = Field(
        default=None,
        title="Node State",
        description="Detailed nodes specific state information",
    )
    reservation: Optional["NodeReservation"] = Field(
        default=None,
        title="Node Reservation",
        description="Information about the current reservation of the node, if any",
    )


class NodeInfo(NodeDefinition):
    """Information about a MADSci Node."""

    actions: dict[str, "ActionDefinition"] = Field(
        title="Node Actions",
        description="The actions that the node supports.",
        default_factory=dict,
    )
    config: Optional[Any] = Field(
        default=None,
        title="Node Configuration",
        description="The current configuration of the node.",
    )
    config_schema: Optional[dict[str, Any]] = Field(
        title="Node Configuration Schema",
        description="JSON Schema for the configuration of the node.",
        default_factory=NodeConfig.model_json_schema,
    )

    @classmethod
    def from_node_def_and_config(
        cls,
        node: NodeDefinition,
        config: Optional[NodeConfig] = None,
    ) -> "NodeInfo":
        """Create a NodeInfo from a NodeDefinition and config."""
        return cls(
            **node.model_dump(exclude={"commands"}),
            config=config,
            config_schema=config.model_json_schema(),
        )


class NodeStatus(BaseModel):
    """Status of a MADSci Node."""

    busy: bool = Field(
        default=False,
        title="Node Busy",
        description="Whether the node is currently at capacity, i.e. running the maximum number of actions allowed.",
    )
    running_actions: set[str] = Field(
        default_factory=set,
        title="Running Action IDs",
        description="The IDs of the actions that the node is currently running.",
    )
    paused: bool = Field(
        default=False,
        title="Node Paused",
        description="Whether the node is paused.",
    )
    locked: bool = Field(
        default=False,
        title="Node Locked",
        description="Whether the node is locked, preventing it from accepting any actions.",
    )
    stopped: bool = Field(
        default=False,
        title="Node Stopped",
        description="Whether the node has been stopped (e.g. due to a safety stop).",
    )
    errored: bool = Field(
        default=False,
        title="Node Errored",
        description="Whether the node is in an errored state.",
    )
    errors: list[Error] = Field(
        default_factory=list,
        title="Node Errors",
        description="A list of errors that the node has encountered.",
    )
    initializing: bool = Field(
        default=False,
        title="Node Initializing",
        description="Whether the node is currently initializing.",
    )
    waiting_for_config: set[str] = Field(
        default_factory=set,
        title="Node Waiting for Configuration",
        description="Set of configuration parameters that the node is waiting for.",
    )
    config_values: dict[str, Any] = Field(
        default_factory=dict,
        title="Node Configuration Values",
        description="The current configuration values of the node.",
    )

    @computed_field
    @property
    def ready(self) -> bool:
        """Whether the node is ready to accept actions."""
        ready = True
        if self.busy:
            ready = False
        if self.locked:
            ready = False
        if self.stopped:
            ready = False
        if self.errored:
            ready = False
        if self.initializing:
            ready = False
        if self.paused:
            ready = False
        if len(self.waiting_for_config) > 0:
            ready = False
        return ready

    @computed_field
    @property
    def description(self) -> str:
        """A description of the node's status."""
        reasons = []
        if self.busy:
            reasons.append("Node is busy")
        if self.locked:
            reasons.append("Node is locked")
        if self.errored:
            reasons.append("Node is in an error state")
        if self.initializing:
            reasons.append("Node is initializing")
        if self.paused:
            reasons.append("Node is paused")
        if len(self.waiting_for_config) > 0:
            reasons.append(
                f"Node is missing configuration values: {self.waiting_for_config}",
            )
        if reasons:
            return "; ".join(reasons)
        return "Node is ready"


class NodeReservation(BaseModel):
    """Reservation of a MADSci Node."""

    owned_by: OwnershipInfo = Field(
        title="Owned By",
        description="Who has ownership of the reservation.",
    )
    created: datetime = Field(
        title="Created Datetime",
        description="When the reservation was created.",
    )
    start: datetime = Field(
        title="Start Datetime",
        description="When the reservation starts.",
    )
    end: datetime = Field(
        title="End Datetime",
        description="When the reservation ends.",
    )

    def check(self, ownership: OwnershipInfo) -> bool:
        """Check if the reservation is 1.) active or not, and 2.) owned by the given ownership."""
        return not (
            not self.owned_by.check(ownership)
            and self.start <= datetime.now()
            and self.end >= datetime.now()
        )


class NodeSetConfigResponse(BaseModel):
    """Response from a Node Set Config Request"""

    success: bool = Field(
        title="Success",
        description="Whether the config was successfully set.",
    )



"""Module to create and manage an object storage client using MinIO."""

import mimetypes
import warnings
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Optional, Union

from madsci.common.types.datapoint_types import ObjectStorageDefinition
from minio import Minio


class ObjectNamingStrategy(Enum):
    """Strategies for naming objects in storage."""

    FILENAME_ONLY = "filename_only"  # Just use the filename
    TIMESTAMPED_PATH = "timestamped_path"  # year/month/day/filename structure


def create_minio_client(
    object_storage_config: ObjectStorageDefinition,
) -> Union[Minio, None]:
    """Initialize the object storage client using the provided configuration."""
    try:
        minio_client = Minio(
            endpoint=object_storage_config.endpoint,
            access_key=object_storage_config.access_key,
            secret_key=object_storage_config.secret_key,
            secure=object_storage_config.secure,
            region=object_storage_config.region
            if object_storage_config.region
            else None,
        )

        try:
            if not ensure_bucket_exists(
                minio_client, object_storage_config.default_bucket
            ):
                minio_client.make_bucket(object_storage_config.default_bucket)

        except Exception as bucket_error:
            # Bucket creation failed - this is OK for many scenarios:
            # - AWS S3: User might not have CreateBucket permissions (bucket created via console)
            # - GCS: Bucket created via GCP console
            # - Bucket already exists but bucket_exists() failed due to permissions
            warnings.warn(
                f"Could not create bucket '{object_storage_config.default_bucket}': {bucket_error!s}. "
                f"Assuming bucket exists and continuing. If uploads fail, please ensure "
                f"the bucket exists and you have appropriate permissions.",
                UserWarning,
                stacklevel=2,
            )

        return minio_client

    except Exception as e:
        warnings.warn(
            f"Failed to initialize object storage client: {e!s}",
            UserWarning,
            stacklevel=2,
        )
        return None


def ensure_bucket_exists(minio_client: Minio, bucket_name: str) -> bool:
    """Ensure a bucket exists, creating it if necessary.

    Args:
        minio_client: The MinIO client instance
        bucket_name: Name of the bucket to check/create

    Returns:
        True if bucket exists or was created successfully, False otherwise
    """
    try:
        if not minio_client.bucket_exists(bucket_name):
            minio_client.make_bucket(bucket_name)
        return True

    except Exception as e:
        warnings.warn(
            f"Failed to check/create bucket '{bucket_name}': {e!s}",
            UserWarning,
            stacklevel=2,
        )
        return False


def get_content_type(file_path: Union[str, Path]) -> str:
    """Get the MIME content type for a file.

    Args:
        file_path: Path to the file

    Returns:
        MIME content type string
    """
    content_type, _ = mimetypes.guess_type(str(file_path))
    return content_type or "application/octet-stream"


def generate_object_name(
    filename: str,
    strategy: ObjectNamingStrategy = ObjectNamingStrategy.FILENAME_ONLY,
    prefix: Optional[str] = None,
) -> str:
    """Generate an object name using the specified strategy.

    Args:
        filename: The original filename
        strategy: Naming strategy to use
        prefix: Optional prefix to add to the object name

    Returns:
        Generated object name
    """
    if strategy == ObjectNamingStrategy.TIMESTAMPED_PATH:
        time = datetime.now()
        base_name = f"{time.year}/{time.month}/{time.day}/{filename}"
    else:  # FILENAME_ONLY
        base_name = filename

    if prefix:
        return f"{prefix}/{base_name}"
    return base_name


def construct_object_url(
    object_storage_config: ObjectStorageDefinition,
    bucket_name: str,
    object_name: str,
    public_endpoint: Optional[str] = None,
) -> str:
    """Construct a URL for accessing an object in storage.

    Args:
        object_storage_config: Object storage configuration
        bucket_name: Name of the bucket
        object_name: Name of the object
        public_endpoint: Optional public endpoint override

    Returns:
        Complete URL to the object
    """
    # Determine the appropriate endpoint for the URL
    if public_endpoint:
        endpoint_for_url = public_endpoint
    else:
        endpoint_for_url = object_storage_config.endpoint
        # If this is a MinIO deployment, use port 9001 for web access instead of 9000
        if ":9000" in endpoint_for_url and (
            "localhost" in endpoint_for_url or "127.0.0.1" in endpoint_for_url
        ):
            # Only adjust for localhost/127.0.0.1 or if explicitly requested
            endpoint_for_url = endpoint_for_url.replace(":9000", ":9001")

    # Construct the object URL
    protocol = "https" if object_storage_config.secure else "http"
    return f"{protocol}://{endpoint_for_url}/{bucket_name}/{object_name}"


def upload_file_to_object_storage(
    minio_client: Minio,
    object_storage_config: ObjectStorageDefinition,
    file_path: Union[str, Path],
    bucket_name: Optional[str] = None,
    object_name: Optional[str] = None,
    content_type: Optional[str] = None,
    metadata: Optional[dict[str, str]] = None,
    naming_strategy: ObjectNamingStrategy = ObjectNamingStrategy.FILENAME_ONLY,
    public_endpoint: Optional[str] = None,
    label: Optional[str] = None,
) -> Optional[dict[str, Any]]:
    """Upload a file to object storage and return storage information.

    Args:
        minio_client: The MinIO client instance
        object_storage_config: Object storage configuration
        file_path: Path to the file to upload
        bucket_name: Name of the bucket (defaults to config default_bucket)
        object_name: Name for the object (auto-generated if not provided)
        content_type: MIME type of the file (auto-detected if not provided)
        metadata: Additional metadata to attach to the object
        naming_strategy: Strategy for generating object names
        public_endpoint: Optional public endpoint for the object storage
        label: Label for the datapoint (defaults to filename)

    Returns:
        Dictionary with object storage information, or None if upload failed
    """
    if minio_client is None:
        warnings.warn(
            "MinIO client is not configured",
            UserWarning,
            stacklevel=2,
        )
        return None

    # Convert to Path object and resolve
    file_path = Path(file_path).expanduser().resolve()

    if not file_path.exists():
        warnings.warn(
            f"File does not exist: {file_path}",
            UserWarning,
            stacklevel=2,
        )
        return None

    # Use defaults if not specified
    bucket_name = bucket_name or object_storage_config.default_bucket
    object_name = object_name or generate_object_name(file_path.name, naming_strategy)
    content_type = content_type or get_content_type(file_path)
    label = label or file_path.name

    # Ensure the bucket exists
    if not ensure_bucket_exists(minio_client, bucket_name):
        return None

    # Upload the file
    try:
        result = minio_client.fput_object(
            bucket_name=bucket_name,
            object_name=object_name,
            file_path=str(file_path),
            content_type=content_type,
            metadata=metadata or {},
        )
    except Exception as e:
        warnings.warn(
            f"Failed to upload file to object storage: {e!s}",
            UserWarning,
            stacklevel=2,
        )
        return None

    # Get file size
    size_bytes = file_path.stat().st_size

    # Construct the object URL
    url = construct_object_url(
        object_storage_config, bucket_name, object_name, public_endpoint
    )

    # Determine public endpoint for response
    final_public_endpoint = public_endpoint
    if not final_public_endpoint:
        final_public_endpoint = object_storage_config.endpoint
        if ":9000" in final_public_endpoint and (
            "localhost" in final_public_endpoint or "127.0.0.1" in final_public_endpoint
        ):
            final_public_endpoint = final_public_endpoint.replace(":9000", ":9001")

    return {
        "bucket_name": bucket_name,
        "object_name": object_name,
        "storage_endpoint": object_storage_config.endpoint,
        "public_endpoint": final_public_endpoint,
        "url": url,
        "label": label,
        "content_type": content_type,
        "size_bytes": size_bytes,
        "etag": result.etag,
        "custom_metadata": metadata or {},
    }


def download_file_from_object_storage(
    minio_client: Minio,
    bucket_name: str,
    object_name: str,
    output_path: Union[str, Path],
) -> bool:
    """Download a file from object storage.

    Args:
        minio_client: The MinIO client instance
        bucket_name: Name of the bucket
        object_name: Name of the object
        output_path: Path where the file should be saved

    Returns:
        True if download was successful, False otherwise
    """
    if minio_client is None:
        warnings.warn(
            "MinIO client is not configured",
            UserWarning,
            stacklevel=2,
        )
        return False

    try:
        output_path = Path(output_path).expanduser()
        output_path.parent.mkdir(parents=True, exist_ok=True)

        minio_client.fget_object(
            bucket_name,
            object_name,
            str(output_path),
        )
        return True

    except Exception as e:
        warnings.warn(
            f"Failed to download from object storage: {e!s}",
            UserWarning,
            stacklevel=2,
        )
        return False


def get_object_data_from_storage(
    minio_client: Minio, bucket_name: str, object_name: str
) -> Optional[bytes]:
    """Get object data directly from storage.

    Args:
        minio_client: The MinIO client instance
        bucket_name: Name of the bucket
        object_name: Name of the object

    Returns:
        Object data as bytes, or None if retrieval failed
    """
    if minio_client is None:
        warnings.warn(
            "MinIO client is not configured",
            UserWarning,
            stacklevel=2,
        )
        return None

    try:
        response = minio_client.get_object(bucket_name, object_name)
        data = response.read()
        response.close()
        response.release_conn()
        return data

    except Exception as e:
        warnings.warn(
            f"Failed to get object from storage: {e!s}",
            UserWarning,
            stacklevel=2,
        )
        return None



"""Types for MADSci Steps."""

from datetime import datetime, timedelta
from typing import Any, Optional, Union

from madsci.common.types.action_types import ActionResult, ActionStatus
from madsci.common.types.base_types import BaseModel, PathLike, new_ulid_str
from madsci.common.types.condition_types import Conditions
from madsci.common.types.location_types import LocationArgument
from sqlmodel.main import Field


class StepDefinition(BaseModel):
    """A definition of a step in a workflow."""

    name: str = Field(
        title="Step Name",
        description="The name of the step.",
    )
    description: Optional[str] = Field(
        title="Step Description",
        description="A description of the step.",
        default=None,
    )
    action: str = Field(
        title="Step Action",
        description="The action to perform in the step.",
    )
    node: str = Field(title="Node Name", description="Name of the node to run on")
    args: dict[str, Any] = Field(
        title="Step Arguments",
        description="Arguments for the step action.",
        default_factory=dict,
    )
    files: dict[str, Optional[PathLike]] = Field(
        title="Step File Arguments",
        description="Files to be used in the step. Key is the name of the file argument, value is the path to the file.",
        default_factory=dict,
    )
    locations: dict[str, Optional[Union[str, LocationArgument]]] = Field(
        title="Step Location Arguments",
        description="Locations to be used in the step. Key is the name of the argument, value is the name of the location, or a Location object.",
        default_factory=dict,
    )
    conditions: list[Conditions] = Field(
        title="Step Conditions",
        description="Conditions for running the step",
        default_factory=list,
    )
    data_labels: dict[str, str] = Field(
        title="Step Data Labels",
        description="Data labels for the results of the step. Maps from the names of the outputs of the action to the names of the data labels.",
        default_factory=dict,
    )


class Step(StepDefinition):
    """A runtime representation of a step in a workflow."""

    step_id: str = Field(
        title="Step ID",
        description="The ID of the step.",
        default_factory=new_ulid_str,
    )
    status: ActionStatus = Field(
        title="Step Status",
        description="The status of the step.",
        default=ActionStatus.NOT_STARTED,
    )
    result: Optional[ActionResult] = Field(
        title="Latest Step Result",
        description="The result of the latest action run.",
        default=None,
    )
    history: list[ActionResult] = Field(
        title="Step History",
        description="The history of the results of the step.",
        default_factory=list,
    )
    start_time: Optional[datetime] = None
    """Time the step started running"""
    end_time: Optional[datetime] = None
    """Time the step finished running"""
    duration: Optional[timedelta] = None
    """Duration of the step's run"""



"""Common validators for MADSci-derived types."""

from typing import Any, Callable, Union

from pydantic import ValidationInfo
from ulid import ULID


def ulid_validator(id: str, info: ValidationInfo) -> str:
    """Validates that a string field is a valid ULID."""
    try:
        ULID.from_str(id)
        return id
    except ValueError as e:
        raise ValueError(f"Invalid ULID {id} for field {info.field_name}") from e


def optional_ulid_validator(id: Union[str, None], info: ValidationInfo) -> str:
    """Validates that a string field is a valid ULID."""
    if id is None:
        return id
    try:
        ULID.from_str(id)
        return id
    except ValueError as e:
        raise ValueError(f"Invalid ULID {id} for field {info.field_name}") from e


def alphanumeric_with_underscores_validator(v: str, info: ValidationInfo) -> str:
    """Validates that a string field is alphanumeric with underscores."""
    if not str(v).replace("_", "").isalnum():
        raise ValueError(
            f"Field {info.field_name} must contain only alphanumeric characters and underscores",
        )
    return v


def create_dict_promoter(
    key_func: Callable[[Any], str],
) -> Callable[[Any], dict[str, Any]]:
    """Creates a validator that promotes a list to a dictionary using a specified attribute as the key.

    Example usage:
        from pydantic import field_validator
        validate_nodes_to_dict = field_validator("nodes", mode="before")(create_dict_promoter("node_name"))
    """

    def list_to_dict(lst: Union[list[Any], dict[str, Any]]) -> dict[str, Any]:
        """Promotes a list to a dictionary using a specified attribute as the key."""
        if isinstance(lst, dict):
            return lst
        return {key_func(item): item for item in lst}

    return list_to_dict



"""Types related to authentication and ownership of MADSci objects."""

from typing import Any, Optional

from madsci.common.types.base_types import BaseModel
from madsci.common.validators import optional_ulid_validator, ulid_validator
from pydantic import SerializationInfo, SerializerFunctionWrapHandler, model_serializer
from pydantic.functional_validators import field_validator
from sqlmodel.main import Field


class OwnershipInfo(BaseModel):
    """Information about the ownership of a MADSci object."""

    user_id: Optional[str] = Field(
        title="User ID",
        description="The ID of the user who owns the object.",
        default=None,
    )
    experiment_id: Optional[str] = Field(
        title="Experiment ID",
        description="The ID of the experiment that owns the object.",
        default=None,
    )
    campaign_id: Optional[str] = Field(
        title="Campaign ID",
        description="The ID of the campaign that owns the object.",
        default=None,
    )
    project_id: Optional[str] = Field(
        title="Project ID",
        description="The ID of the project that owns the object.",
        default=None,
    )
    node_id: Optional[str] = Field(
        title="Node ID",
        description="The ID of the node that owns the object.",
        default=None,
    )
    workcell_id: Optional[str] = Field(
        title="Workcell ID",
        description="The ID of the workcell that owns the object.",
        default=None,
    )
    lab_id: Optional[str] = Field(
        title="Lab ID",
        description="The ID of the lab that owns the object.",
        default=None,
    )
    step_id: Optional[str] = Field(
        title="Step ID",
        description="The ID of the step that owns the object.",
        default=None,
    )
    workflow_id: Optional[str] = Field(
        title="Workflow ID",
        description="The ID of the workflow that owns the object.",
        default=None,
    )
    manager_id: Optional[str] = Field(
        title="Manager ID",
        description="The ID of the manager that owns the object.",
        default=None,
    )

    is_ulid = field_validator(
        "user_id",
        "experiment_id",
        "campaign_id",
        "project_id",
        "node_id",
        "workcell_id",
        "step_id",
        "lab_id",
        "workflow_id",
        "manager_id",
        mode="after",
    )(optional_ulid_validator)

    @model_serializer(mode="wrap")
    def exclude_unset_by_default(
        self, nxt: SerializerFunctionWrapHandler, info: SerializationInfo
    ) -> dict[str, Any]:
        """Exclude unset fields by default."""
        serialized = nxt(self, info)
        return {k: v for k, v in serialized.items() if v is not None}

    def check(self, other: "OwnershipInfo") -> bool:
        """Check if this ownership is the same as another."""
        for key in self.model_dump(exclude_none=True):
            if getattr(self, key) != getattr(other, key):
                return False
        return True


class UserInfo(BaseModel):
    """Information about a user."""

    user_id: str = Field(title="User ID", description="The ID of the user.")
    user_name: str = Field(title="User Name", description="The name of the user.")
    user_email: str = Field(title="User Email", description="The email of the user.")

    is_ulid = field_validator("user_id", mode="after")(ulid_validator)


class ProjectInfo(BaseModel):
    """Information about a project."""

    project_id: str = Field(title="Project ID", description="The ID of the project.")
    project_name: str = Field(
        title="Project Name",
        description="The name of the project.",
    )
    project_description: str = Field(
        title="Project Description",
        description="The description of the project.",
    )
    project_owner: UserInfo = Field(
        title="Project Owner",
        description="The owner of the project.",
    )
    project_members: list[UserInfo] = Field(
        title="Project Members",
        description="The members of the project.",
    )

    is_ulid = field_validator("project_id", mode="after")(ulid_validator)



"""Types for MADSci Worfklow running."""

from datetime import datetime, timedelta
from typing import Any, Optional, Union

from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.base_types import BaseModel, new_ulid_str
from madsci.common.types.step_types import Step, StepDefinition
from madsci.common.validators import ulid_validator
from pydantic import Field, computed_field, field_validator


class WorkflowStatus(BaseModel):
    """Representation of the status of a Workflow"""

    current_step_index: int = 0
    """Index of the current step"""
    paused: bool = False
    """Whether or not the workflow is paused"""
    completed: bool = False
    """Whether or not the workflow has completed successfully"""
    failed: bool = False
    """Whether or not the workflow has failed"""
    cancelled: bool = False
    """Whether or not the workflow has been cancelled"""
    running: bool = False
    """Whether or not the workflow is currently running"""
    has_started: bool = False
    """Whether or not at least one step of the workflow has been run"""

    def reset(self, step_index: int = 0) -> None:
        """Reset the workflow status"""
        self.current_step_index = step_index
        self.paused = False
        self.completed = False
        self.failed = False
        self.cancelled = False
        self.running = False
        self.has_started = False

    @computed_field
    @property
    def queued(self) -> bool:
        """Whether or not the workflow is queued"""
        return self.active and not self.running

    @computed_field
    @property
    def active(self) -> bool:
        """Whether or not the workflow is actively being scheduled"""
        return not (self.terminal or self.paused)

    @computed_field
    @property
    def terminal(self) -> bool:
        """Whether or not the workflow is in a terminal state"""
        return self.completed or self.failed or self.cancelled

    @computed_field
    @property
    def started(self) -> bool:
        """Whether or not the workflow has started"""
        return self.current_step_index > 0

    @computed_field
    @property
    def ok(self) -> bool:
        """Whether or not the workflow is ok (i.e. not failed or cancelled)"""
        return not (self.failed or self.cancelled)

    @computed_field
    @property
    def description(self) -> str:  # noqa: PLR0911
        """Description of the workflow's status"""
        if self.completed:
            return "Completed Successfully"
        if self.failed:
            return f"Failed on step {self.current_step_index}"
        if self.cancelled:
            return f"Cancelled on step {self.current_step_index}"
        if self.paused:
            return f"Paused on step {self.current_step_index}"
        if self.running:
            return f"Running step {self.current_step_index}"
        if self.active:
            return f"Queued on step {self.current_step_index}"
        return "Unknown"


class WorkflowParameter(BaseModel):
    """container for a workflow parameter"""

    name: str
    """the name of the parameter"""
    default: Optional[Any] = None
    """ the default value of the parameter"""


class WorkflowMetadata(BaseModel, extra="allow"):
    """Metadata container"""

    author: Optional[str] = None
    """Who wrote this object"""
    description: Optional[str] = None
    """Description of the object"""
    version: Union[float, str] = ""
    """Version of the object"""


class WorkflowDefinition(BaseModel):
    """Grand container that pulls all info of a workflow together"""

    name: str
    """Name of the workflow"""
    workflow_metadata: WorkflowMetadata = Field(default_factory=WorkflowMetadata)
    """Information about the flow"""
    parameters: Optional[list[WorkflowParameter]] = Field(default_factory=list)
    """Inputs to the workflow"""
    steps: list[StepDefinition] = Field(default_factory=list)
    """User Submitted Steps of the flow"""

    @field_validator("steps", mode="after")
    @classmethod
    def ensure_data_label_uniqueness(cls, v: Any) -> Any:
        """Ensure that the names of the arguments and files are unique"""
        labels = []
        for step in v:
            if step.data_labels:
                for key in step.data_labels:
                    if step.data_labels[key] in labels:
                        raise ValueError("Data labels must be unique across workflow")
                    labels.append(step.data_labels[key])
        return v


class SchedulerMetadata(BaseModel):
    """Scheduler information"""

    ready_to_run: bool = False
    """Whether or not the next step in the workflow is ready to run"""
    priority: int = 0
    """Used to rank workflows when deciding which to run next. Higher is more important"""
    reasons: list[str] = Field(default_factory=list)
    """Allow the scheduler to provide reasons for its decisions"""


class Workflow(WorkflowDefinition):
    """Container for a workflow run"""

    scheduler_metadata: SchedulerMetadata = Field(default_factory=SchedulerMetadata)
    """scheduler information for the workflow run"""
    label: Optional[str] = None
    """Label for the workflow run"""
    workflow_id: str = Field(default_factory=new_ulid_str)
    """ID of the workflow run"""
    steps: list[Step] = Field(default_factory=list)
    """Processed Steps of the flow"""
    parameter_values: dict[str, Any] = Field(default_factory=dict)
    """parameter values used in this workflow"""
    ownership_info: OwnershipInfo = Field(default_factory=OwnershipInfo)
    """ID of the experiment this workflow is a part of"""
    status: WorkflowStatus = Field(default_factory=WorkflowStatus)
    """current status of the workflow"""
    step_index: int = 0
    """Index of the current step"""
    simulate: bool = False
    """Whether or not this workflow is being simulated"""
    submitted_time: Optional[datetime] = None
    """Time workflow was submitted to the scheduler"""
    start_time: Optional[datetime] = None
    """Time the workflow started running"""
    end_time: Optional[datetime] = None
    """Time the workflow finished running"""
    duration: Optional[timedelta] = None
    """Duration of the workflow's run"""
    step_definitions: list[StepDefinition] = Field(default_factory=list)
    """The original step definitions for the workflow"""

    def get_step_by_name(self, name: str) -> Step:
        """Return the step object by its name"""
        for step in self.steps:
            if step.name == name:
                return step
        raise KeyError(f"Step {name} not found in workflow run {self.run_id}")

    def get_step_by_id(self, id: str) -> Step:
        """Return the step object indexed by its id"""
        for step in self.steps:
            if step.id == id:
                return step
        raise KeyError(f"Step {id} not found in workflow run {self.run_id}")

    def get_datapoint_id_by_label(self, label: str) -> str:
        """Return the ID of the first datapoint with the given label in a workflow run"""
        for step in self.steps:
            if step.result.data:
                for key in step.result.data:
                    if key == label:
                        return step.result.data[key]
        raise KeyError(f"Label {label} not found in workflow run {self.run_id}")

    def get_all_datapoint_ids_by_label(self, label: str) -> list[str]:
        """Return the IDs of all datapoints with the given label in a workflow run"""
        ids = []
        for step in self.steps:
            if step.result.data:
                for key in step.result.data:
                    if key == label:
                        ids.append(step.result.data[key])
        if not ids:
            raise KeyError(f"Label {label} not found in workflow run {self.run_id}")
        return ids

    is_ulid = field_validator("workflow_id")(ulid_validator)



"""Types used by the Resource Manager's Server"""

from typing import Optional, Union

from madsci.common.types.auth_types import OwnershipInfo
from madsci.common.types.base_types import BaseModel
from madsci.common.types.resource_types import (
    GridIndex,
    GridIndex2D,
    GridIndex3D,
    ResourceDataModels,
)
from pydantic import model_validator
from pydantic.config import ConfigDict
from pydantic.types import datetime


class ResourceRequestBase(BaseModel):
    """Base class for all resource request models."""

    model_config = ConfigDict(
        extra="forbid",
    )


class ResourceGetQuery(ResourceRequestBase):
    """A request to get a resource from the database."""

    resource_id: Optional[str] = None
    """The ID of the resource"""
    resource_name: Optional[str] = None
    """The name of the resource."""
    resource_description: Optional[str] = None
    """The description of the resource."""
    parent_id: Optional[str] = None
    """The ID of the parent resource"""
    resource_class: Optional[str] = None
    """The class of the resource."""
    base_type: Optional[str] = None
    """The base type of the resource"""
    owner: Optional[OwnershipInfo] = None
    """The owner(s) of the resource"""
    unique: Optional[bool] = False
    """Whether to require a unique resource or not."""
    multiple: Optional[bool] = True
    """Whether to return multiple resources or just the first."""


class ResourceHistoryGetQuery(ResourceRequestBase):
    """A request to get the history of a resource from the database."""

    resource_id: Optional[str] = None
    """The ID of the resource."""
    version: Optional[int] = None
    """The version of the resource."""
    removed: Optional[bool] = None
    """Whether the resource was removed."""
    change_type: Optional[str] = None
    """The type of change to the resource."""
    start_date: Optional[datetime] = None
    """The start a range from which to get history. If not specified, all history before the end date is returned."""
    end_date: Optional[datetime] = None
    """The end of a range from which to get history. If not specified, all history after the start date is returned."""
    limit: Optional[int] = None
    """The maximum number of entries to return."""


class PushResourceBody(ResourceRequestBase):
    """A request to push a resource to the database."""

    child_id: Optional[str] = None
    """The ID of the child resource."""
    child: Optional[ResourceDataModels] = None
    """The child resource data."""

    @model_validator(mode="before")
    @classmethod
    def validate_push_resource(cls, values: dict) -> dict:
        """Ensure that either a child ID or child resource data is provided."""
        if not values.get("child_id") and not values.get("child"):
            raise ValueError(
                "Either a child ID or child resource data must be provided."
            )
        return values


class SetChildBody(ResourceRequestBase):
    """A request to set a child resource."""

    key: Union[str, GridIndex, GridIndex2D, GridIndex3D]
    """The key to identify the child resource's location in the parent container. If the parent is a grid/voxel grid, the key should be a 2D or 3D index."""
    child: Union[str, ResourceDataModels]
    """The ID of the child resource or the child resource data."""


class RemoveChildBody(ResourceRequestBase):
    """A request to remove a child resource."""

    key: Union[str, GridIndex2D, GridIndex3D]
    """The key to identify the child resource's location in the parent container. If the parent is a grid/voxel grid, the key should be a 2D or 3D index."""



"""Exceptions common across the MADSci Framework"""


class ActionMissingArgumentError(ValueError):
    """An action was requested with a missing argument"""


class ActionMissingFileError(ValueError):
    """An action was requested with a missing file argument"""


class ActionNotImplementedError(ValueError):
    """An action was requested, but isn't implemented by the node"""


class WorkflowFailedError(Exception):
    """Raised when a workflow fails"""

    def __init__(self, message: str) -> "WorkflowFailedError":
        """Initializes the exception"""
        super().__init__(message)
        self.message = message


class WorkflowCanceledError(Exception):
    """Raised when a workflow is canceled"""

    def __init__(self, message: str) -> "WorkflowCanceledError":
        """Initializes the exception"""
        super().__init__(message)
        self.message = message


class ExperimentCancelledError(Exception):
    """Raised in an experiment application when an experiment is cancelled"""

    def __init__(self, message: str) -> "ExperimentCancelledError":
        """Initializes the exception"""
        super().__init__(message)
        self.message = message


class ExperimentFailedError(Exception):
    """Raised in an experiment application when an experiment fails."""

    def __init__(self, message: str) -> "ExperimentCancelledError":
        """Initializes the exception"""
        super().__init__(message)
        self.message = message
